import uuid

import django_filters
import taggit
from django import forms
from django.conf import settings as django_settings
from django.contrib import auth
from django.contrib.contenttypes.models import ContentType
from django.core import exceptions
from django.db.models import Q
from django.db.models.functions import Concat
from django.utils import timezone
from django_filters.filterset import FilterSetMetaclass
from django_filters.rest_framework import DjangoFilterBackend
from django_filters.widgets import BooleanWidget
from rest_framework.filters import BaseFilterBackend

from waldur_core.core import filters as core_filters
from waldur_core.core import models as core_models
from waldur_core.core.filters import ExternalFilterBackend
from waldur_core.core.utils import order_with_nulls, get_ordering, is_uuid_like
from waldur_core.structure import SupportedServices
from waldur_core.structure import models
from waldur_core.structure.managers import filter_queryset_for_user

User = auth.get_user_model()


class NameFilterSet(django_filters.FilterSet):
    name = django_filters.CharFilter(lookup_expr='icontains')
    name_exact = django_filters.CharFilter(field_name='name', lookup_expr='exact')


class ScopeTypeFilterBackend(DjangoFilterBackend):
    """ Scope filters:

        * ?scope = ``URL``
        * ?scope_type = ``string`` (can be list)
    """

    content_type_field = 'content_type'
    scope_param = 'scope_type'
    scope_models = {
        'customer': models.Customer,
        'service': models.Service,
        'project': models.Project,
        'service_project_link': models.ServiceProjectLink,
        'resource': models.ResourceMixin
    }

    @classmethod
    def get_scope_type(cls, model):
        for scope_type, scope_model in cls.scope_models.items():
            if issubclass(model, scope_model):
                return scope_type

    @classmethod
    def _get_scope_models(cls, types):
        for scope_type, scope_model in cls.scope_models.items():
            if scope_type in types:
                try:
                    for submodel in scope_model.get_all_models():
                        yield submodel
                except AttributeError:
                    yield scope_model

    @classmethod
    def _get_scope_content_types(cls, types):
        return ContentType.objects.get_for_models(*cls._get_scope_models(types)).values()

    def filter_queryset(self, request, queryset, view):
        if self.scope_param in request.query_params:
            content_types = self._get_scope_content_types(request.query_params.getlist(self.scope_param))
            return queryset.filter(**{'%s__in' % self.content_type_field: content_types})
        return queryset


class GenericRoleFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        return filter_queryset_for_user(queryset, request.user)


class GenericUserFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        user_uuid = request.query_params.get('user_uuid')
        if not user_uuid:
            return queryset

        try:
            uuid.UUID(user_uuid)
        except ValueError:
            return queryset.none()

        try:
            user = User.objects.get(uuid=user_uuid)
        except User.DoesNotExist:
            return queryset.none()

        return filter_queryset_for_user(queryset, user)


class CustomerFilter(NameFilterSet):
    query = django_filters.CharFilter(method='filter_query')
    native_name = django_filters.CharFilter(lookup_expr='icontains')
    abbreviation = django_filters.CharFilter(lookup_expr='icontains')
    contact_details = django_filters.CharFilter(lookup_expr='icontains')
    division_uuid = django_filters.UUIDFilter(field_name='division__uuid')
    division_name = django_filters.CharFilter(field_name='division__name', lookup_expr='icontains')

    class Meta:
        model = models.Customer
        fields = [
            'name',
            'abbreviation',
            'contact_details',
            'native_name',
            'registration_code',
            'agreement_number',
            'backend_id',
        ]

    def filter_query(self, queryset, name, value):
        if value:
            return queryset.filter(
                Q(name__icontains=value) |
                Q(native_name__icontains=value) |
                Q(abbreviation__icontains=value) |
                Q(domain__icontains=value) |
                Q(uuid__icontains=value) |
                Q(registration_code__icontains=value) |
                Q(agreement_number__contains=value)
            )
        return queryset


class ExternalCustomerFilterBackend(ExternalFilterBackend):
    pass


class AccountingStartDateFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        query = Q(accounting_start_date__gt=timezone.now())
        return filter_by_accounting_is_running(request, queryset, query)


class CustomerAccountingStartDateFilter(BaseFilterBackend):
    def filter_queryset(self, request, queryset, view):
        query = Q(customer__accounting_start_date__gt=timezone.now())
        return filter_by_accounting_is_running(request, queryset, query)


def filter_by_accounting_is_running(request, queryset, query):
    if not django_settings.WALDUR_CORE['ENABLE_ACCOUNTING_START_DATE']:
        return queryset

    value = request.query_params.get('accounting_is_running')
    boolean_field = forms.NullBooleanField()

    try:
        value = boolean_field.to_python(value)
    except exceptions.ValidationError:
        value = None

    if value is None:
        return queryset

    if value:
        return queryset.exclude(query)
    else:
        return queryset.filter(query)


class ProjectTypeFilter(NameFilterSet):

    class Meta:
        model = models.ProjectType
        fields = ['name']


class ProjectFilter(NameFilterSet):
    customer = django_filters.UUIDFilter(
        field_name='customer__uuid',
        distinct=True,
    )

    customer_name = django_filters.CharFilter(
        field_name='customer__name',
        distinct=True,
        lookup_expr='icontains'
    )

    customer_native_name = django_filters.CharFilter(
        field_name='customer__native_name',
        distinct=True,
        lookup_expr='icontains'
    )

    customer_abbreviation = django_filters.CharFilter(
        field_name='customer__abbreviation',
        distinct=True,
        lookup_expr='icontains'
    )

    description = django_filters.CharFilter(lookup_expr='icontains')

    query = django_filters.CharFilter(method='filter_query')

    o = django_filters.OrderingFilter(
        fields=(
            ('name', 'name'),
            ('created', 'created'),
            ('customer__name', 'customer_name'),
            ('customer__native_name', 'customer_native_name'),
            ('customer__abbreviation', 'customer_abbreviation'),
        )
    )

    class Meta:
        model = models.Project
        fields = [
            'name',
            'customer', 'customer_name', 'customer_native_name', 'customer_abbreviation',
            'description',
            'created',
            'query',
        ]

    def filter_query(self, queryset, name, value):
        if is_uuid_like(value):
            return queryset.filter(uuid=value)
        else:
            return queryset.filter(name__icontains=value)


class CustomerUserFilter(DjangoFilterBackend):
    def filter_queryset(self, request, queryset, view):
        customer_uuid = request.query_params.get('customer_uuid')
        if not customer_uuid:
            return queryset

        try:
            uuid.UUID(customer_uuid)
        except ValueError:
            return queryset.none()

        return queryset.filter(
            Q(customerpermission__customer__uuid=customer_uuid,
              customerpermission__is_active=True) |
            Q(projectpermission__project__customer__uuid=customer_uuid,
              projectpermission__is_active=True)
        ).distinct()


class ProjectUserFilter(DjangoFilterBackend):
    def filter_queryset(self, request, queryset, view):
        project_uuid = request.query_params.get('project_uuid')
        if not project_uuid:
            return queryset

        try:
            uuid.UUID(project_uuid)
        except ValueError:
            return queryset.none()

        return queryset.filter(
            projectpermission__project__uuid=project_uuid,
            projectpermission__is_active=True
        ).distinct()


def filter_visible_users(queryset, user, extra=None):
    connected_customers_query = models.Customer.objects.all()
    if not (user.is_staff or user.is_support):
        connected_customers_query = connected_customers_query.filter(
            Q(permissions__user=user, permissions__is_active=True) |
            Q(projects__permissions__user=user, projects__permissions__is_active=True)
        ).distinct()

    connected_customers = list(connected_customers_query.all())

    subquery = (
        Q(customerpermission__customer__in=connected_customers,
          customerpermission__is_active=True) |
        Q(projectpermission__project__customer__in=connected_customers,
          projectpermission__is_active=True)
    )

    queryset = queryset.filter(is_staff=False).filter(
        subquery | Q(uuid=user.uuid) | (extra or Q())
    ).distinct()

    if not (user.is_staff or user.is_support):
        queryset = queryset.filter(is_active=True, is_staff=False)

    return queryset


class UserFilterBackend(DjangoFilterBackend):
    def filter_queryset(self, request, queryset, view):
        user = request.user

        current_user = request.query_params.get('current')
        if current_user is not None and not user.is_anonymous:
            queryset = User.objects.filter(uuid=user.uuid)

        if (not django_settings.WALDUR_CORE.get('SHOW_ALL_USERS', False) and
                not (user.is_staff or user.is_support)):
            queryset = filter_visible_users(queryset, user, self.get_extra_q(user))

        return queryset.order_by('username')

    _extra_query = []

    @classmethod
    def register_extra_query(cls, func_get_query):
        """
        Add extra Q for user list queryset
        :param func_get_query: a function that takes User object and returns Q object
        :return: None
        """
        cls._extra_query.append(func_get_query)

    @classmethod
    def get_extra_q(cls, user):
        result = Q()
        for q in cls._extra_query:
            result = result | q(user)
        return result


class BaseUserFilter(django_filters.FilterSet):
    full_name = django_filters.CharFilter(lookup_expr='icontains')
    username = django_filters.CharFilter()
    native_name = django_filters.CharFilter(lookup_expr='icontains')
    organization = django_filters.CharFilter(lookup_expr='icontains')
    job_title = django_filters.CharFilter(lookup_expr='icontains')
    email = django_filters.CharFilter(lookup_expr='icontains')
    is_active = django_filters.BooleanFilter(widget=BooleanWidget)

    class Meta:
        model = User
        fields = [
            'full_name',
            'native_name',
            'organization',
            'email',
            'phone_number',
            'description',
            'job_title',
            'username',
            'civil_number',
            'is_active',
            'registration_method',
        ]


class UserFilter(BaseUserFilter):
    is_staff = django_filters.BooleanFilter(widget=BooleanWidget)
    is_support = django_filters.BooleanFilter(widget=BooleanWidget)

    o = django_filters.OrderingFilter(
        fields=('full_name', 'native_name', 'email', 'phone_number',
                'description', 'organization', 'job_title', 'username',
                'is_active', 'registration_method', 'is_staff', 'is_support')
    )


class UserConcatenatedNameOrderingBackend(DjangoFilterBackend):
    """ Filter user by concatenated full_name + username with ?o=concatenated_name """

    def filter_queryset(self, request, queryset, view):
        queryset = self._filter_queryset(request, queryset, view)
        return BaseUserFilter(request.query_params, queryset=queryset, request=request).qs

    def _filter_queryset(self, request, queryset, view):
        if 'o' not in request.query_params:
            return queryset
        if request.query_params['o'] == 'concatenated_name':
            order_by = 'concatenated_name'
        elif request.query_params['o'] == '-concatenated_name':
            order_by = '-concatenated_name'
        else:
            return queryset
        return queryset.annotate(concatenated_name=Concat('full_name', 'username')).order_by(order_by)


class UserPermissionFilter(django_filters.FilterSet):
    user = django_filters.UUIDFilter(field_name='user__uuid')
    user_url = core_filters.URLFilter(
        view_name='user-detail',
        field_name='user__uuid',
    )
    username = django_filters.CharFilter(
        field_name='user__username',
        lookup_expr='exact',
    )
    full_name = django_filters.CharFilter(
        field_name='user__full_name',
        lookup_expr='icontains',
    )
    native_name = django_filters.CharFilter(
        field_name='user__native_name',
        lookup_expr='icontains',
    )

    o = django_filters.OrderingFilter(
        fields=(
            ('user__username', 'username'),
            ('user__full_name', 'full_name'),
            ('user__native_name', 'native_name'),
            ('user__email', 'email'),
            ('expiration_time', 'expiration_time'),
            ('created', 'created'),
            ('role', 'role'),
        )
    )


class ProjectPermissionFilter(UserPermissionFilter):
    class Meta:
        fields = ['role']
        model = models.ProjectPermission

    customer = django_filters.UUIDFilter(
        field_name='project__customer__uuid',
    )
    project = django_filters.UUIDFilter(
        field_name='project__uuid',
    )
    project_url = core_filters.URLFilter(
        view_name='project-detail',
        field_name='project__uuid',
    )


class CustomerPermissionFilter(UserPermissionFilter):
    class Meta:
        fields = ['role']
        model = models.CustomerPermission

    customer = django_filters.UUIDFilter(
        field_name='customer__uuid',
    )
    customer_url = core_filters.URLFilter(
        view_name='customer-detail',
        field_name='customer__uuid',
    )


class SshKeyFilter(NameFilterSet):
    uuid = django_filters.UUIDFilter()
    user_uuid = django_filters.UUIDFilter(field_name='user__uuid')

    o = django_filters.OrderingFilter(fields=('name',))

    class Meta:
        model = core_models.SshPublicKey
        fields = [
            'name',
            'fingerprint',
            'uuid',
            'user_uuid',
            'is_shared',
        ]


class ServiceTypeFilter(django_filters.Filter):
    def filter(self, qs, value):
        value = SupportedServices.get_filter_mapping().get(value)
        return super(ServiceTypeFilter, self).filter(qs, value)


class ServiceSettingsFilter(NameFilterSet):
    type = ServiceTypeFilter()
    state = core_filters.StateFilter()
    has_resources = django_filters.BooleanFilter(method='filter_has_resources', widget=BooleanWidget)

    class Meta:
        model = models.ServiceSettings
        fields = ('name', 'type', 'state', 'shared')

    def filter_has_resources(self, queryset, name, value):
        service_settings_with_resources = []
        for model in SupportedServices.get_resource_models().values():
            service_settings_with_resources \
                += list(model.objects
                        .values_list('service_project_link__service__settings__pk', flat=True))
        if value is True:
            return queryset.filter(pk__in=service_settings_with_resources)
        return queryset.exclude(Q(pk__in=service_settings_with_resources))


class ServiceSettingsScopeFilterBackend(core_filters.GenericKeyFilterBackend):
    def get_related_models(self):
        return models.ResourceMixin.get_all_models()

    def get_field_name(self):
        return 'scope'


class ServiceFilterMetaclass(FilterSetMetaclass):
    """ Build a list of supported resource via serializers definition.
        See SupportedServices for details.
    """

    def __new__(mcs, name, bases, args):
        service_filter = super(ServiceFilterMetaclass, mcs).__new__(mcs, name, bases, args)
        model = args['Meta'].model
        if not model._meta.abstract:
            SupportedServices.register_service_filter(args['Meta'].model, service_filter)
        return service_filter


class BaseServiceFilter(django_filters.FilterSet, metaclass=ServiceFilterMetaclass):
    customer = django_filters.UUIDFilter(field_name='customer__uuid')
    name = django_filters.CharFilter(field_name='settings__name', lookup_expr='icontains')
    name_exact = django_filters.CharFilter(field_name='settings__name', lookup_expr='exact')
    project = core_filters.URLFilter(view_name='project-detail', field_name='projects__uuid', distinct=True)
    project_uuid = django_filters.UUIDFilter(field_name='projects__uuid', distinct=True)
    settings = core_filters.URLFilter(view_name='servicesettings-detail', field_name='settings__uuid', distinct=True)
    shared = django_filters.BooleanFilter(field_name='settings__shared', distinct=True, widget=BooleanWidget)
    type = ServiceTypeFilter(field_name='settings__type')
    tag = django_filters.ModelMultipleChoiceFilter(
        field_name='settings__tags__name',
        to_field_name='name',
        lookup_expr='in',
        queryset=taggit.models.Tag.objects.all(),
    )
    # rtag - required tag, support for filtration by tags using AND operation
    # ?rtag=t1&rtag=t2 - will filter instances that have both t1 and t2.
    rtag = django_filters.ModelMultipleChoiceFilter(
        field_name='settings__tags__name',
        to_field_name='name',
        queryset=taggit.models.Tag.objects.all(),
        conjoined=True,
    )

    class Meta:
        model = models.Service
        fields = ('name', 'name_exact', 'project_uuid',
                  'customer', 'project', 'settings', 'shared', 'type', 'tag', 'rtag')


class BaseServiceProjectLinkFilter(django_filters.FilterSet):
    service_uuid = django_filters.UUIDFilter(field_name='service__uuid')
    settings_uuid = django_filters.UUIDFilter(field_name='service__settings__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='service__customer__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='project__uuid')
    project = core_filters.URLFilter(view_name='project-detail', field_name='project__uuid')

    class Meta:
        model = models.ServiceProjectLink
        fields = ()


class ResourceFilterMetaclass(FilterSetMetaclass):
    """ Build a list of supported resource via serializers definition.
        See SupportedServices for details.
    """

    def __new__(cls, name, bases, args):
        resource_filter = super(ResourceFilterMetaclass, cls).__new__(cls, name, bases, args)
        if 'Meta' in args:
            SupportedServices.register_resource_filter(args['Meta'].model, resource_filter)
        return resource_filter


class BaseResourceFilter(NameFilterSet, metaclass=ResourceFilterMetaclass):
    def __init__(self, *args, **kwargs):
        super(BaseResourceFilter, self).__init__(*args, **kwargs)
        self.filters['o'] = django_filters.OrderingFilter(fields=self.ORDERING_FIELDS)

    # customer
    customer = django_filters.UUIDFilter(field_name='service_project_link__service__customer__uuid')
    customer_uuid = django_filters.UUIDFilter(field_name='service_project_link__service__customer__uuid')
    customer_name = django_filters.CharFilter(
        field_name='service_project_link__service__customer__name', lookup_expr='icontains')
    customer_native_name = django_filters.CharFilter(
        field_name='service_project_link__project__customer__native_name', lookup_expr='icontains')
    customer_abbreviation = django_filters.CharFilter(
        field_name='service_project_link__project__customer__abbreviation', lookup_expr='icontains')
    # project
    project = django_filters.UUIDFilter(field_name='service_project_link__project__uuid')
    project_uuid = django_filters.UUIDFilter(field_name='service_project_link__project__uuid')
    project_name = django_filters.CharFilter(field_name='service_project_link__project__name', lookup_expr='icontains')
    # service
    service_uuid = django_filters.UUIDFilter(field_name='service_project_link__service__uuid')
    service_name = django_filters.CharFilter(field_name='service_project_link__service__settings__name',
                                             lookup_expr='icontains')
    # service settings
    service_settings_uuid = django_filters.UUIDFilter(field_name='service_project_link__service__settings__uuid')
    service_settings_name = django_filters.CharFilter(field_name='service_project_link__service__settings__name',
                                                      lookup_expr='icontains')
    # resource
    description = django_filters.CharFilter(lookup_expr='icontains')
    state = core_filters.MappedMultipleChoiceFilter(
        choices=[(representation, representation) for db_value, representation in
                 core_models.StateMixin.States.CHOICES],
        choice_mappings={representation: db_value for db_value, representation in
                         core_models.StateMixin.States.CHOICES},
    )
    uuid = django_filters.UUIDFilter(lookup_expr='exact')
    backend_id = django_filters.CharFilter(field_name='backend_id', lookup_expr='exact')
    tag = django_filters.ModelMultipleChoiceFilter(
        field_name='tags__name',
        label='tag',
        to_field_name='name',
        queryset=taggit.models.Tag.objects.all(),
    )
    rtag = django_filters.ModelMultipleChoiceFilter(
        field_name='tags__name',
        label='rtag',
        to_field_name='name',
        queryset=taggit.models.Tag.objects.all(),
        conjoined=True,
    )
    external_ip = core_filters.EmptyFilter()

    ORDERING_FIELDS = (
        ('name', 'name'),
        ('state', 'state'),
        ('service_project_link__project__customer__name', 'customer_name'),
        ('service_project_link__project__customer__native_name', 'customer_native_name'),
        ('service_project_link__project__customer__abbreviation', 'customer_abbreviation'),
        ('service_project_link__project__name', 'project_name'),
        ('service_project_link__service__settings__name', 'service_name'),
        ('service_project_link__service__uuid', 'service_uuid'),
        ('created', 'created'),
    )

    class Meta:
        model = models.ResourceMixin
        fields = (
            # customer
            'customer', 'customer_uuid', 'customer_name', 'customer_native_name', 'customer_abbreviation',
            # project
            'project', 'project_uuid', 'project_name',
            # service
            'service_uuid', 'service_name',
            # service settings
            'service_settings_name', 'service_settings_uuid',
            # resource
            'name', 'name_exact', 'description', 'state', 'uuid', 'backend_id', 'tag', 'rtag',
        )


class TagsFilter(BaseFilterBackend):
    """ Tags ordering. Filtering for complex tags.

    Example:
        ?tag__license-os=centos7 - will filter objects with tag "license-os:centos7".

    Allow to define next parameters in view:
     - tags_filter_db_field - name of tags field in database. Default: tags.
     - tags_filter_request_field - name of tags in request. Default: tag.
    """

    def filter_queryset(self, request, queryset, view):
        self.db_field = getattr(view, 'tags_filter_db_field', 'tags')
        self.request_field = getattr(view, 'tags_filter_request_field', 'tag')

        queryset = self._filter(request, queryset)
        queryset = self._order(request, queryset)
        return queryset

    def _filter(self, request, queryset):
        for key in request.query_params.keys():
            item_name = self._get_item_name(key)
            if item_name:
                value = request.query_params.get(key)
                filter_kwargs = {
                    self.db_field + '__name__startswith': item_name,
                    self.db_field + '__name__icontains': value,
                }
                queryset = queryset.filter(**filter_kwargs)
        return queryset

    def _order(self, request, queryset):
        order_by = get_ordering(request)
        item_name = self._get_item_name(order_by)
        if item_name:
            filter_kwargs = {self.db_field + '__name__startswith': item_name}
            queryset = queryset.filter(**filter_kwargs).order_by(self.db_field + '__name')
        return queryset

    def _get_item_name(self, key):
        prefix = self.request_field + '__'
        if key and key.startswith(prefix):
            return key[len(prefix):]


class StartTimeFilter(BaseFilterBackend):

    def filter_queryset(self, request, queryset, view):
        order_by = get_ordering(request)
        if order_by not in ('start_time', '-start_time'):
            return queryset
        return order_with_nulls(queryset, order_by)


class BaseServicePropertyFilter(NameFilterSet):

    class Meta:
        fields = ('name', 'name_exact')


class ServicePropertySettingsFilter(BaseServicePropertyFilter):
    settings_uuid = django_filters.UUIDFilter(field_name='settings__uuid')
    settings = core_filters.URLFilter(view_name='servicesettings-detail', field_name='settings__uuid', distinct=True)

    class Meta(BaseServicePropertyFilter.Meta):
        fields = BaseServicePropertyFilter.Meta.fields + ('settings_uuid', 'settings')


class ResourceSummaryFilterBackend(core_filters.SummaryFilter):
    """ Filter and order SummaryQuerySet of resources """

    def get_queryset_filter(self, queryset):
        try:
            return SupportedServices.get_resource_filter(queryset.model)
        except KeyError:
            return super(ResourceSummaryFilterBackend, self).get_queryset_filter(queryset)

    def get_base_filter(self):
        return BaseResourceFilter


class ServiceSummaryFilterBackend(core_filters.SummaryFilter):
    def get_queryset_filter(self, queryset):
        try:
            return SupportedServices.get_service_filter(queryset.model)
        except KeyError:
            return super(ServiceSummaryFilterBackend, self).get_queryset_filter(queryset)

    def get_base_filter(self):
        return BaseServiceFilter


class DivisionFilter(NameFilterSet):
    type = django_filters.CharFilter(field_name='type__name', lookup_expr='iexact')
    parent = django_filters.UUIDFilter(field_name='parent__uuid')

    class Meta:
        model = models.Division
        fields = [
            'name',
        ]



from azure.common.credentials import ServicePrincipalCredentials
from azure.mgmt.compute import ComputeManagementClient
from azure.mgmt.compute.models import DiskCreateOption, LinuxConfiguration, \
    OSProfile, SshConfiguration, SshPublicKey
from azure.mgmt.network import NetworkManagementClient
from azure.mgmt.network.models import NetworkInterface, NetworkInterfaceIPConfiguration, \
    NetworkSecurityGroup, SecurityRule
from azure.mgmt.rdbms.postgresql import PostgreSQLManagementClient
from azure.mgmt.rdbms.postgresql.models import ServerForCreate, \
    ServerPropertiesForDefaultCreate, ServerVersion, StorageProfile
from azure.mgmt.resource import ResourceManagementClient, SubscriptionClient
from azure.mgmt.storage import StorageManagementClient
from django.utils.functional import cached_property
from msrest.exceptions import ClientException

from waldur_core.structure import ServiceBackendError


class AzureBackendError(ServiceBackendError):
    pass


class AzureClient:
    def __init__(self, settings):
        self.subscription_id = str(settings.options['subscription_id'])
        self.client_id = str(settings.options['client_id'])
        self.client_secret = str(settings.options['client_secret'])
        self.tenant_id = str(settings.options['tenant_id'])

    @cached_property
    def credentials(self):
        return ServicePrincipalCredentials(
            client_id=self.client_id,
            secret=self.client_secret,
            tenant=self.tenant_id,
        )

    @cached_property
    def subscription_client(self):
        return SubscriptionClient(self.credentials)

    @cached_property
    def resource_client(self):
        return ResourceManagementClient(self.credentials, self.subscription_id)

    @cached_property
    def compute_client(self):
        return ComputeManagementClient(self.credentials, self.subscription_id, '2018-06-01')

    @cached_property
    def storage_client(self):
        return StorageManagementClient(self.credentials, self.subscription_id)

    @cached_property
    def network_client(self):
        return NetworkManagementClient(self.credentials, self.subscription_id)

    @cached_property
    def pgsql_client(self):
        return PostgreSQLManagementClient(self.credentials, self.subscription_id)

    def list_locations(self):
        try:
            return self.subscription_client.subscriptions.list_locations(self.subscription_id)
        except ClientException as exc:
            raise AzureBackendError(exc)

    def get_resource_group_locations(self):
        """
        Resource Manager is supported in all regions, but the resources
        you deploy might not be supported in all regions.
        In addition, there may be limitations on your subscription that
        prevent you from using some regions that support the resource.

        See also: https://docs.microsoft.com/en-us/azure/azure-resource-manager/resource-manager-supported-services
        """
        try:
            provider = self.resource_client.providers.get('Microsoft.Resources')
        except ClientException as exc:
            raise AzureBackendError(exc)
        else:
            for resource in provider.resource_types:
                if resource.resource_type == 'resourceGroups':
                    return resource.locations

    def list_resource_groups(self):
        try:
            return self.resource_client.resource_groups.list()
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_resource_group(self, location, resource_group_name):
        try:
            return self.resource_client.resource_groups.create_or_update(
                resource_group_name,
                {'location': location}
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def delete_resource_group(self, resource_group_name):
        try:
            return self.resource_client.resource_groups.delete(resource_group_name)
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_virtual_machine_sizes(self, location):
        try:
            return self.compute_client.virtual_machine_sizes.list(location)
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_virtual_machine_images(self, location):
        try:
            publishers = self.compute_client.virtual_machine_images.list_publishers(location)

            for publisher in publishers:
                offers = self.compute_client.virtual_machine_images.list_offers(
                    location,
                    publisher.name,
                )

                for offer in offers:
                    skus = self.compute_client.virtual_machine_images.list_skus(
                        location,
                        publisher.name,
                        offer.name,
                    )

                    for sku in skus:
                        result_list = self.compute_client.virtual_machine_images.list(
                            location,
                            publisher.name,
                            offer.name,
                            sku.name,
                        )

                        for version in result_list:
                            yield self.compute_client.virtual_machine_images.get(
                                location,
                                publisher.name,
                                offer.name,
                                sku.name,
                                version.name,
                            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_all_virtual_machines(self):
        try:
            return self.compute_client.virtual_machines.list_all()
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_virtual_machines_in_group(self, resource_group_name):
        try:
            return self.compute_client.virtual_machines.list(resource_group_name)
        except ClientException as exc:
            raise AzureBackendError(exc)

    def get_virtual_machine(self, resource_group_name, vm_name):
        try:
            return self.compute_client.virtual_machines.get(
                resource_group_name,
                vm_name
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_virtual_machine(self, location, resource_group_name,
                               vm_name, size_name, nic_id, image_reference,
                               username, password, custom_data=None, ssh_key=None):
        os_profile = OSProfile(
            computer_name=vm_name,
            admin_username=username,
            admin_password=password,
        )
        if custom_data:
            os_profile.custom_data = custom_data

        if ssh_key:
            os_profile.linux_configuration = LinuxConfiguration(
                ssh=SshConfiguration(
                    public_keys=[
                        SshPublicKey(key_data=ssh_key),
                    ],
                )
            )
        try:
            return self.compute_client.virtual_machines.create_or_update(
                resource_group_name,
                vm_name,
                {
                    'location': location,
                    'os_profile': os_profile,
                    'hardware_profile': {
                        'vm_size': size_name,
                    },
                    'storage_profile': {
                        'image_reference': image_reference,
                    },
                    'network_profile': {
                        'network_interfaces': [{
                            'id': nic_id,
                        }]
                    },
                },
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def delete_virtual_machine(self, resource_group_name, vm_name):
        try:
            return self.compute_client.virtual_machines.delete(
                resource_group_name,
                vm_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def start_virtual_machine(self, resource_group_name, vm_name):
        try:
            return self.compute_client.virtual_machines.start(
                resource_group_name,
                vm_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def restart_virtual_machine(self, resource_group_name, vm_name):
        try:
            return self.compute_client.virtual_machines.restart(
                resource_group_name,
                vm_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def stop_virtual_machine(self, resource_group_name, vm_name):
        try:
            return self.compute_client.virtual_machines.power_off(
                resource_group_name,
                vm_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_storage_account(self, location, resource_group_name, account_name):
        try:
            return self.storage_client.storage_accounts.create(
                resource_group_name,
                account_name,
                {
                    'sku': {'name': 'standard_lrs'},
                    'kind': 'storage',
                    'location': location
                }
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_disk(self, location, resource_group_name, disk_name, disk_size_gb):
        try:
            return self.compute_client.disks.create_or_update(
                resource_group_name,
                disk_name,
                {
                    'location': location,
                    'disk_size_gb': disk_size_gb,
                    'creation_data': {
                        'create_option': DiskCreateOption.empty
                    }
                }
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_network(self, location, resource_group_name, network_name, cidr):
        try:
            return self.network_client.virtual_networks.create_or_update(
                resource_group_name,
                network_name,
                {
                    'location': location,
                    'address_space': {
                        'address_prefixes': [cidr]
                    }
                }
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_subnet(self, resource_group_name, network_name, subnet_name, cidr):
        try:
            return self.network_client.subnets.create_or_update(
                resource_group_name,
                network_name,
                subnet_name,
                {
                    'address_prefix': cidr,
                }
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def get_network_interface(self, resource_group_name, network_interface_name):
        try:
            return self.network_client.network_interfaces.get(
                resource_group_name,
                network_interface_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_network_interface(self, location, resource_group_name,
                                 interface_name, config_name, subnet_id,
                                 public_ip_id=None, security_group_id=None):
        ip_configuration = NetworkInterfaceIPConfiguration(
            name=config_name,
            subnet={'id': subnet_id}
        )

        if public_ip_id:
            ip_configuration.public_ip_address = {'id': public_ip_id}

        interface_parameters = NetworkInterface(
            location=location,
            ip_configurations=[ip_configuration],
        )

        if security_group_id:
            interface_parameters.network_security_group = {'id': security_group_id}

        try:
            return self.network_client.network_interfaces.create_or_update(
                resource_group_name,
                interface_name,
                interface_parameters
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_ssh_security_group(self, location, resource_group_name, network_security_group_name):
        ssh_rule = SecurityRule(
            name='default-allow-ssh',
            protocol='Tcp',
            source_port_range='*',
            destination_port_range=22,
            direction='Inbound',
            source_address_prefix='*',
            destination_address_prefix='*',
            access='Allow',
            priority=1000,
        )

        security_group = NetworkSecurityGroup(
            location=location,
            security_rules=[ssh_rule]
        )

        try:
            return self.network_client.network_security_groups.create_or_update(
                resource_group_name,
                network_security_group_name,
                security_group,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def get_public_ip(self, resource_group_name, public_ip_address_name):
        try:
            return self.network_client.public_ip_addresses.get(
                resource_group_name,
                public_ip_address_name
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_all_public_ips(self):
        try:
            return self.network_client.public_ip_addresses.list_all()
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_public_ip(self, location, resource_group_name, public_ip_address_name):
        try:
            return self.network_client.public_ip_addresses.create_or_update(
                resource_group_name,
                public_ip_address_name,
                {
                    'location': location
                }
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def delete_public_ip(self, resource_group_name, public_ip_address_name):
        try:
            return self.network_client.public_ip_addresses.delete(
                resource_group_name,
                public_ip_address_name
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_all_sql_servers(self):
        try:
            return self.pgsql_client.servers.list()
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_sql_servers_in_group(self, resource_group_name):
        try:
            return self.pgsql_client.servers.list_by_resource_group(resource_group_name)
        except ClientException as exc:
            raise AzureBackendError(exc)

    def get_sql_server(self, resource_group_name, server_name):
        try:
            return self.pgsql_client.servers.get_by_resource_group(
                resource_group_name,
                server_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_sql_server(self, location, resource_group_name, server_name,
                          username, password, sku=None, storage_mb=None, ssl_enforcement=None):
        properties = ServerPropertiesForDefaultCreate(
            administrator_login=username,
            administrator_login_password=password,
            version=ServerVersion.nine_full_stop_six,
            ssl_enforcement=ssl_enforcement,
        )
        if storage_mb:
            properties.storage_profile = StorageProfile(storage_mb=storage_mb)
        try:
            poller = self.pgsql_client.servers.create(
                resource_group_name,
                server_name,
                ServerForCreate(
                    properties=properties,
                    location=location,
                    sku=sku,
                )
            )
            return poller.result()
        except ClientException as exc:
            raise AzureBackendError(exc)

    def delete_sql_server(self, resource_group_name, server_name):
        try:
            return self.pgsql_client.servers.delete(resource_group_name, server_name)
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_sql_firewall_rule(self, resource_group_name, server_name, firewall_rule_name,
                                 start_ip_address, end_ip_address):
        try:
            poller = self.pgsql_client.firewall_rules.create_or_update(
                resource_group_name,
                server_name,
                firewall_rule_name,
                start_ip_address,
                end_ip_address,
            )
            return poller.result()
        except ClientException as exc:
            raise AzureBackendError(exc)

    def get_sql_database(self, resource_group_name, server_name, database_name):
        try:
            return self.pgsql_client.databases.get(
                resource_group_name,
                server_name,
                database_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def create_sql_database(self, resource_group_name, server_name, database_name,
                            charset=None, collation=None):
        try:
            return self.pgsql_client.databases.create_or_update(
                resource_group_name,
                server_name,
                database_name,
                charset,
                collation,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def list_sql_databases_in_server(self, resource_group_name, server_name):
        try:
            return self.pgsql_client.databases.list_by_server(
                resource_group_name,
                server_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)

    def delete_sql_database(self, resource_group_name, server_name, database_name):
        try:
            return self.pgsql_client.databases.delete(
                resource_group_name,
                server_name,
                database_name,
            )
        except ClientException as exc:
            raise AzureBackendError(exc)



import itertools
import json
import logging

from django import forms
from django.conf import settings
from django.conf.urls import url
from django.contrib import admin, messages
from django.contrib.admin import SimpleListFilter
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.core.exceptions import ValidationError, FieldDoesNotExist
from django.db import models as django_models
from django.forms import ModelMultipleChoiceField, ModelForm, RadioSelect, ChoiceField, CharField
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _
from django.utils.translation import ungettext

from waldur_core.core import utils as core_utils
from waldur_core.core.admin import (
    get_admin_url, ExecutorAdminAction, PasswordWidget,
    NativeNameAdminMixin, JsonWidget, ReadOnlyAdminMixin,
)
from waldur_core.core.admin_filters import RelatedOnlyDropdownFilter
from waldur_core.core.models import User
from waldur_core.core.validators import BackendURLValidator
from waldur_core.quotas.admin import QuotaInline
from waldur_core.structure import models, SupportedServices, executors, utils
from waldur_geo_ip import tasks as geo_ip_tasks

logger = logging.getLogger(__name__)


class BackendModelAdmin(admin.ModelAdmin):

    def get_list_filter(self, request):
        try:
            self.model._meta.get_field('settings')
            return ('settings__shared', ('settings', RelatedOnlyDropdownFilter))
        except FieldDoesNotExist:
            return self.list_filter

    def lookup_allowed(self, lookup, value):
        if lookup == 'settings__shared__exact':
            return True
        return super(BackendModelAdmin, self).lookup_allowed(lookup, value)

    def has_add_permission(self, request):
        return False

    def get_readonly_fields(self, request, obj=None):
        fields = super(BackendModelAdmin, self).get_readonly_fields(request, obj)

        if not obj:
            return fields

        excluded = self.get_exclude(request, obj) or tuple()
        if not settings.WALDUR_CORE['BACKEND_FIELDS_EDITABLE']:
            instance_class = type(obj)
            fields = fields + instance_class.get_backend_fields()
            fields = filter(lambda field: field not in excluded, fields)

        return fields


class FormRequestAdminMixin:
    """
    This mixin allows you to get current request user in the model admin form,
    which then passed to add_user method, so that user which granted role,
    is stored in the permission model.
    """

    def get_form(self, request, obj=None, **kwargs):
        form = super(FormRequestAdminMixin, self).get_form(request, obj=obj, **kwargs)
        form.request = request
        return form


class ChangeReadonlyMixin:
    add_readonly_fields = ()
    change_readonly_fields = ()

    def get_readonly_fields(self, request, obj=None):
        fields = super(ChangeReadonlyMixin, self).get_readonly_fields(request, obj)
        if hasattr(request, '_is_admin_add_view') and request._is_admin_add_view:
            return tuple(set(fields) | set(self.add_readonly_fields))
        else:
            return tuple(set(fields) | set(self.change_readonly_fields))

    def add_view(self, request, *args, **kwargs):
        request._is_admin_add_view = True
        return super(ChangeReadonlyMixin, self).add_view(request, *args, **kwargs)


class ProtectedModelMixin:
    def delete_view(self, request, *args, **kwargs):
        try:
            response = super(ProtectedModelMixin, self).delete_view(request, *args, **kwargs)
        except django_models.ProtectedError as e:
            self.message_user(request, e, messages.ERROR)
            return HttpResponseRedirect('.')
        else:
            return response


class ResourceCounterFormMixin:
    def get_vm_count(self, obj):
        return obj.quotas.get(name=obj.Quotas.nc_vm_count).usage

    get_vm_count.short_description = _('VM count')

    def get_app_count(self, obj):
        return obj.quotas.get(name=obj.Quotas.nc_app_count).usage

    get_app_count.short_description = _('Application count')

    def get_private_cloud_count(self, obj):
        return obj.quotas.get(name=obj.Quotas.nc_private_cloud_count).usage

    get_private_cloud_count.short_description = _('Private cloud count')


class CustomerAdminForm(ModelForm):
    owners = ModelMultipleChoiceField(User.objects.all().order_by('full_name'), required=False,
                                      widget=FilteredSelectMultiple(verbose_name=_('Owners'), is_stacked=False))
    support_users = ModelMultipleChoiceField(User.objects.all().order_by('full_name'), required=False,
                                             widget=FilteredSelectMultiple(verbose_name=_('Support users'),
                                                                           is_stacked=False))

    def __init__(self, *args, **kwargs):
        super(CustomerAdminForm, self).__init__(*args, **kwargs)
        if self.instance and self.instance.pk:
            self.owners = self.instance.get_owners()
            self.support_users = self.instance.get_support_users()
            self.fields['owners'].initial = self.owners
            self.fields['support_users'].initial = self.support_users
        else:
            self.owners = User.objects.none()
            self.support_users = User.objects.none()
        self.fields['agreement_number'].initial = models.get_next_agreement_number()

        textarea_attrs = {'cols': '40', 'rows': '4'}
        self.fields['contact_details'].widget.attrs = textarea_attrs
        self.fields['access_subnets'].widget.attrs = textarea_attrs
        type_choices = ['']
        type_choices.extend(settings.WALDUR_CORE['COMPANY_TYPES'])
        self.fields['type'] = ChoiceField(required=False, choices=[(t, t) for t in type_choices])

    def save(self, commit=True):
        customer = super(CustomerAdminForm, self).save(commit=False)

        if not customer.pk:
            customer.save()

        self.populate_users('owners', customer, models.CustomerRole.OWNER)
        self.populate_users('support_users', customer, models.CustomerRole.SUPPORT)

        return customer

    def populate_users(self, field_name, customer, role):
        field = getattr(self, field_name)
        new_users = self.cleaned_data[field_name]

        removed_users = field.exclude(pk__in=new_users)
        for user in removed_users:
            customer.remove_user(user, role, self.request.user)

        added_users = new_users.exclude(pk__in=field)
        for user in added_users:
            # User role within customer must be unique.
            if not customer.has_user(user):
                customer.add_user(user, role, self.request.user)

        self.save_m2m()

    def clean(self):
        cleaned_data = super(CustomerAdminForm, self).clean()
        owners = self.cleaned_data['owners']
        support_users = self.cleaned_data['support_users']
        invalid_users = set(owners) & set(support_users)
        if invalid_users:
            invalid_users_list = ', '.join(map(str, invalid_users))
            raise ValidationError(_('User role within organization must be unique. '
                                    'Role assignment of The following users is invalid: %s.') % invalid_users_list)
        return cleaned_data

    def clean_accounting_start_date(self):
        accounting_start_date = self.cleaned_data['accounting_start_date']
        if 'accounting_start_date' in self.changed_data and accounting_start_date < timezone.now():
            # If accounting_start_date < timezone.now(), we change accounting_start_date
            # but not raise an exception, because accounting_start_date default value is
            # timezone.now(), but init time of form and submit time of form are always diff.
            # And user will get an exception always if set default value.
            return timezone.now()

        return accounting_start_date


class CustomerAdmin(FormRequestAdminMixin,
                    ResourceCounterFormMixin,
                    NativeNameAdminMixin,
                    ProtectedModelMixin,
                    admin.ModelAdmin):
    form = CustomerAdminForm
    fields = ('name', 'uuid', 'image', 'native_name', 'abbreviation', 'division', 'contact_details',
              'registration_code', 'backend_id', 'domain',
              'agreement_number', 'email', 'phone_number', 'access_subnets', 'homepage',
              'country', 'vat_code', 'is_company', 'owners', 'support_users',
              'type', 'address', 'postal', 'latitude', 'longitude',
              'bank_name', 'bank_account',
              'accounting_start_date', 'default_tax_percent', 'blocked')
    list_display = ('name', 'uuid', 'abbreviation',
                    'created', 'accounting_start_date',
                    'get_vm_count', 'get_app_count', 'get_private_cloud_count')
    list_filter = ('blocked', 'division')
    search_fields = ('name', 'uuid', 'abbreviation')
    date_hierarchy = 'created'
    readonly_fields = ('uuid',)
    inlines = [QuotaInline]

    def get_readonly_fields(self, request, obj=None):
        fields = super(CustomerAdmin, self).get_readonly_fields(request, obj)
        if obj and obj.is_billable():
            return fields + ('accounting_start_date',)
        return fields


class ProjectAdminForm(ModelForm):
    admins = ModelMultipleChoiceField(User.objects.all().order_by('full_name'), required=False,
                                      widget=FilteredSelectMultiple(verbose_name=_('Admins'), is_stacked=False))
    managers = ModelMultipleChoiceField(User.objects.all().order_by('full_name'), required=False,
                                        widget=FilteredSelectMultiple(verbose_name=_('Managers'), is_stacked=False))
    support_users = ModelMultipleChoiceField(User.objects.all().order_by('full_name'), required=False,
                                             widget=FilteredSelectMultiple(verbose_name=_('Support users'),
                                                                           is_stacked=False))

    def __init__(self, *args, **kwargs):
        super(ProjectAdminForm, self).__init__(*args, **kwargs)
        if self.instance and self.instance.pk:
            self.admins = self.instance.get_users(models.ProjectRole.ADMINISTRATOR)
            self.managers = self.instance.get_users(models.ProjectRole.MANAGER)
            self.support_users = self.instance.get_users(models.ProjectRole.SUPPORT)
            self.fields['admins'].initial = self.admins
            self.fields['managers'].initial = self.managers
            self.fields['support_users'].initial = self.support_users
        else:
            for field_name in ('admins', 'managers', 'support_users'):
                setattr(self, field_name, User.objects.none())

    def clean(self):
        cleaned_data = super(ProjectAdminForm, self).clean()
        admins = self.cleaned_data['admins']
        managers = self.cleaned_data['managers']
        support_users = self.cleaned_data['support_users']
        for xs, ys in itertools.combinations([set(admins), set(managers), set(support_users)], 2):
            invalid_users = xs & ys
            if invalid_users:
                invalid_users_list = ', '.join(map(str, invalid_users))
                raise ValidationError(_('User role within project must be unique. '
                                        'Role assignment of The following users is invalid: %s.') % invalid_users_list)
        return cleaned_data

    def save(self, commit=True):
        project = super(ProjectAdminForm, self).save(commit=False)

        if not project.pk:
            project.save()

        self.populate_users('admins', project, models.ProjectRole.ADMINISTRATOR)
        self.populate_users('managers', project, models.ProjectRole.MANAGER)
        self.populate_users('support_users', project, models.ProjectRole.SUPPORT)

        return project

    def populate_users(self, field_name, project, role):
        field = getattr(self, field_name)
        new_users = self.cleaned_data[field_name]

        removed_users = field.exclude(pk__in=new_users)
        for user in removed_users:
            project.remove_user(user, role, self.request.user)

        added_users = new_users.exclude(pk__in=field)
        for user in added_users:
            # User role within project must be unique.
            if not project.has_user(user):
                project.add_user(user, role, self.request.user)
        self.save_m2m()


class ProjectAdmin(FormRequestAdminMixin,
                   ResourceCounterFormMixin,
                   ProtectedModelMixin,
                   ChangeReadonlyMixin,
                   admin.ModelAdmin):
    form = ProjectAdminForm

    fields = ('name', 'description', 'customer', 'type',
              'admins', 'managers', 'support_users', 'certifications')

    list_display = ['name', 'uuid', 'customer', 'created', 'get_type_name',
                    'get_vm_count', 'get_app_count', 'get_private_cloud_count']
    search_fields = ['name', 'uuid']
    change_readonly_fields = ['customer']
    inlines = [QuotaInline]
    filter_horizontal = ('certifications',)
    actions = ('cleanup',)

    class Cleanup(ExecutorAdminAction):
        executor = executors.ProjectCleanupExecutor
        short_description = _('Delete projects with all resources')

    cleanup = Cleanup()

    def get_type_name(self, project):
        return project.type and project.type.name or ''

    get_type_name.short_description = _('Type')
    get_type_name.admin_order_field = 'type__name'


class ServiceCertificationAdmin(admin.ModelAdmin):
    list_display = ('name', 'link')
    search_fields = ['name', 'link']
    list_filter = ('service_settings',)


class ServiceSettingsAdminForm(ModelForm):
    backend_url = CharField(max_length=200, required=False, validators=[BackendURLValidator()])

    def clean(self):
        cleaned_data = super(ServiceSettingsAdminForm, self).clean()
        service_type = cleaned_data.get('type')
        if not service_type:
            return

        field_info = utils.get_all_services_field_info()
        fields_required = field_info.fields_required
        extra_fields_required = field_info.extra_fields_required
        fields_default = field_info.extra_fields_default[service_type]

        # Check required fields of service type
        for field in fields_required[service_type]:
            value = cleaned_data.get(field)
            if not value:
                try:
                    self.add_error(field, _('This field is required.'))
                except ValueError:
                    logger.warning('Incorrect field %s in %s required_fields' %
                                   (field, service_type))

        # Check required extra fields of service type
        try:
            if 'options' in cleaned_data:
                options = json.loads(cleaned_data.get('options'))
                unfilled = set(extra_fields_required[service_type]) - set(options.keys()) - set(fields_default.keys())

                if unfilled:
                    self.add_error('options', _('This field must include keys: %s') %
                                   ', '.join(unfilled))
                service_serializer = SupportedServices.get_service_serializer_for_key(service_type)
                options_serializer_class = getattr(service_serializer.Meta, 'options_serializer', None)
                if options_serializer_class:
                    options_serializer = options_serializer_class(data=options)
                    if not options_serializer.is_valid():
                        self.add_error('options', json.dumps(options_serializer.errors))
                    else:
                        cleaned_data['options'] = options_serializer.validated_data
        except ValueError:
            self.add_error('options', _('JSON is not valid'))

        return cleaned_data

    class Meta:
        widgets = {
            'options': JsonWidget(),
            'geolocations': JsonWidget(),
            'username': forms.TextInput(attrs={'autocomplete': 'new-password'}),
            'password': PasswordWidget(attrs={'autocomplete': 'new-password'}),
            'token': PasswordWidget(attrs={'autocomplete': 'new-password'}),
        }

    def __init__(self, *args, **kwargs):
        super(ServiceSettingsAdminForm, self).__init__(*args, **kwargs)
        self.fields['type'] = ChoiceField(choices=SupportedServices.get_choices(),
                                          widget=RadioSelect)


class ServiceTypeFilter(SimpleListFilter):
    title = 'type'
    parameter_name = 'type'

    def lookups(self, request, model_admin):
        return SupportedServices.get_choices()

    def queryset(self, request, queryset):
        if self.value():
            return queryset.filter(type=self.value())
        else:
            return queryset


class PrivateServiceSettingsAdmin(ChangeReadonlyMixin, admin.ModelAdmin):
    readonly_fields = ('error_message', 'uuid')
    list_display = ('name', 'customer', 'get_type_display', 'state', 'error_message', 'uuid')
    list_filter = (ServiceTypeFilter, 'state')
    search_fields = ('name', 'uuid')
    change_readonly_fields = ('customer',)
    actions = ['pull']
    form = ServiceSettingsAdminForm
    fields = ('type', 'name', 'uuid', 'backend_url', 'username', 'password',
              'token', 'domain', 'certificate', 'options', 'customer',
              'state', 'error_message', 'tags', 'homepage', 'terms_of_services',
              'certifications', 'geolocations')
    inlines = [QuotaInline]
    filter_horizontal = ('certifications',)
    common_fields = ('type', 'name', 'uuid', 'options', 'customer',
                     'state', 'error_message', 'tags', 'homepage', 'terms_of_services',
                     'certifications', 'geolocations')

    # must be specified explicitly not to be constructed from model name by default.
    change_form_template = 'admin/structure/servicesettings/change_form.html'

    def get_type_display(self, obj):
        return obj.get_type_display()

    get_type_display.short_description = 'Type'

    def add_view(self, *args, **kwargs):
        self.exclude = getattr(self, 'add_exclude', ())
        return super(PrivateServiceSettingsAdmin, self).add_view(*args, **kwargs)

    def changeform_view(self, request, object_id=None, form_url='', extra_context=None):
        extra_context = extra_context or {}
        field_info = utils.get_all_services_field_info()
        service_field_names = field_info.fields
        service_fields_required = field_info.fields_required

        for service_name in service_field_names:
            service_field_names[service_name].extend(self.common_fields)

        extra_context['service_fields'] = json.dumps(service_field_names)
        extra_context['service_fields_required'] = json.dumps(service_fields_required)
        return super(PrivateServiceSettingsAdmin, self).changeform_view(request, object_id, form_url, extra_context)

    def get_readonly_fields(self, request, obj=None):
        fields = super(PrivateServiceSettingsAdmin, self).get_readonly_fields(request, obj)
        if not obj:
            return fields + ('state',)
        elif obj.scope:
            return fields + ('options',)
        return fields

    def get_urls(self):
        my_urls = [
            url(r'^(.+)/change/services/$', self.admin_site.admin_view(self.services)),
        ]
        return my_urls + super(PrivateServiceSettingsAdmin, self).get_urls()

    def services(self, request, pk=None):
        settings = models.ServiceSettings.objects.get(id=pk)
        projects = {}

        spl_model = SupportedServices.get_related_models(settings)['service_project_link']
        for spl in spl_model.objects.filter(service__settings=settings):
            projects.setdefault(spl.project.id, {
                'name': str(spl.project),
                'url': get_admin_url(spl.project),
                'services': [],
            })
            projects[spl.project.id]['services'].append({
                'name': str(spl.service),
                'url': get_admin_url(spl.service),
            })

        return render(request, 'structure/service_settings_entities.html', {'projects': projects.values()})

    class Pull(ExecutorAdminAction):
        executor = executors.ServiceSettingsPullExecutor
        short_description = _('Pull')

        def validate(self, service_settings):
            States = models.ServiceSettings.States
            if service_settings.state not in (States.OK, States.ERRED):
                raise ValidationError(_('Service settings has to be OK or erred.'))

    pull = Pull()

    def save_model(self, request, obj, form, change):
        obj.save()
        if not change:
            executors.ServiceSettingsCreateExecutor.execute(obj)


class SharedServiceSettingsAdmin(PrivateServiceSettingsAdmin):
    actions = ['pull', 'connect_shared']

    def get_fields(self, request, obj=None):
        fields = super(SharedServiceSettingsAdmin, self).get_fields(request, obj)
        return [field for field in fields if field != 'customer']

    def get_list_display(self, request):
        fields = super(SharedServiceSettingsAdmin, self).get_list_display(request)
        return [field for field in fields if field != 'customer']

    def save_form(self, request, form, change):
        obj = super(SharedServiceSettingsAdmin, self).save_form(request, form, change)

        """If required field is not filled, but it has got a default value, we set a default value."""
        field_info = utils.get_all_services_field_info()
        extra_fields_default = field_info.extra_fields_default[obj.type]
        extra_fields_required = field_info.extra_fields_required[obj.type]
        default = (set(extra_fields_required) - set(obj.options.keys())) & set(extra_fields_default.keys())

        if default:
            for d in default:
                obj.options[d] = extra_fields_default[d]

        if not change:
            obj.shared = True
        return obj

    class ConnectShared(ExecutorAdminAction):
        executor = executors.ServiceSettingsConnectSharedExecutor
        short_description = _('Create SPLs and services for shared service settings')

        def validate(self, service_settings):
            if not service_settings.shared:
                raise ValidationError(_('It is impossible to connect not shared settings.'))

    connect_shared = ConnectShared()


class ServiceAdmin(admin.ModelAdmin):
    list_display = ('settings', 'customer')
    list_filter = (
        ('settings', RelatedOnlyDropdownFilter),
        ('customer', RelatedOnlyDropdownFilter),
    )
    ordering = ('customer',)


class ServicePropertyAdmin(ReadOnlyAdminMixin, admin.ModelAdmin):
    search_fields = ('name',)
    list_filter = (
        ('settings', RelatedOnlyDropdownFilter),
    )
    readonly_fields = ('name', 'settings')
    list_display = ('name', 'settings')


class ServiceProjectLinkAdmin(admin.ModelAdmin):
    list_display = ('get_service_name', 'get_customer_name', 'get_project_name')
    list_filter = ('service__settings', 'project__name', 'service__settings__name')
    ordering = ('service__customer__name', 'project__name')
    list_display_links = ('get_service_name',)
    search_fields = ('service__customer__name', 'project__name', 'service__settings__name')
    inlines = [QuotaInline]

    def get_readonly_fields(self, request, obj=None):
        if obj:
            return ('service', 'project')
        return ()

    def get_queryset(self, request):
        queryset = super(ServiceProjectLinkAdmin, self).get_queryset(request)
        return queryset.select_related('service', 'project', 'project__customer')

    def get_service_name(self, obj):
        return obj.service.settings.name

    get_service_name.short_description = _('Service')

    def get_project_name(self, obj):
        return obj.project.name

    get_project_name.short_description = _('Project')

    def get_customer_name(self, obj):
        return obj.service.customer.name

    get_customer_name.short_description = _('Customer')


class DerivedFromSharedSettingsResourceFilter(SimpleListFilter):
    title = _('service settings')
    parameter_name = 'shared__exact'

    def lookups(self, request, model_admin):
        return ((1, _('Shared')), (0, _('Private')))

    def queryset(self, request, queryset):
        if self.value() is not None:
            return queryset.filter(service_project_link__service__settings__shared=self.value())
        else:
            return queryset


class ResourceAdmin(BackendModelAdmin):
    readonly_fields = ('error_message',)
    list_display = ('uuid', 'name', 'backend_id', 'state', 'created',
                    'get_service', 'get_project', 'error_message', 'get_settings_shared')
    list_filter = BackendModelAdmin.list_filter + ('state', DerivedFromSharedSettingsResourceFilter)
    search_fields = ('name',)

    def get_settings_shared(self, obj):
        return obj.service_project_link.service.settings.shared

    get_settings_shared.short_description = _('Are service settings shared')

    def get_service(self, obj):
        return obj.service_project_link.service

    get_service.short_description = _('Service')
    get_service.admin_order_field = 'service_project_link__service__settings__name'

    def get_project(self, obj):
        return obj.service_project_link.project

    get_project.short_description = _('Project')
    get_project.admin_order_field = 'service_project_link__project__name'


class PublishableResourceAdmin(ResourceAdmin):
    list_display = ResourceAdmin.list_display + ('publishing_state',)


class VirtualMachineAdmin(ResourceAdmin):
    readonly_fields = ResourceAdmin.readonly_fields + ('image_name',)

    actions = ['detect_coordinates']

    def detect_coordinates(self, request, queryset):
        geo_ip_tasks.detect_vm_coordinates_batch.delay([core_utils.serialize_instance(vm) for vm in queryset])
        tasks_scheduled = queryset.count()
        message = ungettext(
            'Coordinates detection has been scheduled for one virtual machine.',
            'Coordinates detection has been scheduled for %(tasks_scheduled)d virtual machines.',
            tasks_scheduled
        )
        message = message % {'tasks_scheduled': tasks_scheduled}

        self.message_user(request, message)

    detect_coordinates.short_description = _('Detect coordinates of virtual machines')


class DivisionTypeAdmin(admin.ModelAdmin):
    list_display = ('name',)
    search_fields = ['name']


class DivisionAdmin(admin.ModelAdmin):
    list_display = ('name', 'type', 'parent')
    search_fields = ['name']
    list_filter = ('type',)


admin.site.register(models.ServiceCertification, ServiceCertificationAdmin)
admin.site.register(models.Customer, CustomerAdmin)
admin.site.register(models.ProjectType, admin.ModelAdmin)
admin.site.register(models.Project, ProjectAdmin)
admin.site.register(models.PrivateServiceSettings, PrivateServiceSettingsAdmin)
admin.site.register(models.SharedServiceSettings, SharedServiceSettingsAdmin)
admin.site.register(models.DivisionType, DivisionTypeAdmin)
admin.site.register(models.Division, DivisionAdmin)



from django.apps import AppConfig


class BookingConfig(AppConfig):
    name = 'waldur_mastermind.booking'
    verbose_name = 'Booking system'

    def ready(self):
        from waldur_mastermind.marketplace.plugins import manager

        from . import PLUGIN_NAME, processors

        manager.register(offering_type=PLUGIN_NAME,
                         create_resource_processor=processors.BookingCreateProcessor,
                         delete_resource_processor=processors.BookingDeleteProcessor)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-04-10 07:17
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0001_squashed_0008'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='user',
            name='organization_approved',
        ),
    ]



from datetime import datetime

from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers

from waldur_core.core import utils as core_utils


class StringTimestampField(serializers.CharField):
    """
    This field serializes datetime string representation to the timestamp (e.g. '2016-08-11T10:33:38' -> 1470911618).
    Note that there must be at least one format provided in formats parameter (e.g. formats=('%Y-%m-%dT%H:%M:%S',)).
    First format from formats list will be used during deserialization process.
    """

    def __init__(self, formats=(), **kwargs):
        assert formats, 'At least one datetime string format must be provided.'
        self.formats = formats
        super(StringTimestampField, self).__init__(**kwargs)

    def to_representation(self, value):
        for str_format in self.formats:
            try:
                date_time = datetime.strptime(value, str_format)
            except ValueError:
                pass
            else:
                return core_utils.datetime_to_timestamp(date_time)

        raise serializers.ValidationError(_('This field does not have datetime format that matches %s string.') % value)

    def to_internal_value(self, value):
        try:
            date_time = core_utils.timestamp_to_datetime(value)
            datetime_str = date_time.strftime(self.formats[0])
        except ValueError:
            raise serializers.ValidationError(_('Value "{}" should be valid UNIX timestamp.').format(value))
        return datetime_str



import datetime
import logging

from django.contrib.contenttypes.models import ContentType
from django.core import exceptions as django_exceptions
from django.db.models import Q

from waldur_core.core import models as core_models
from waldur_core.core import utils as core_utils
from waldur_core.core.utils import month_start
from waldur_mastermind.common import mixins as common_mixins
from waldur_mastermind.invoices import models as invoice_models
from waldur_mastermind.invoices import registrators
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace.plugins import manager
from waldur_mastermind.marketplace_rancher import PLUGIN_NAME

logger = logging.getLogger(__name__)


def create_usage(cluster):
    try:
        resource = marketplace_models.Resource.objects.get(scope=cluster)
    except django_exceptions.ObjectDoesNotExist:
        logger.debug('Skipping node usage synchronization because this '
                     'marketplace.Resource does not exist.'
                     'Cluster ID: %s', cluster.id)
        return

    date = datetime.date.today()
    usage = cluster.node_set.filter(state=core_models.StateMixin.States.OK).count()

    for component in manager.get_components(PLUGIN_NAME):
        try:
            offering_component = marketplace_models.OfferingComponent.objects.get(
                offering=resource.offering,
                type=component.type
            )
            plan_period = marketplace_models.ResourcePlanPeriod.objects. \
                filter(Q(start__lte=date) | Q(start__isnull=True)). \
                filter(Q(end__gt=date) | Q(end__isnull=True)). \
                get(resource=resource)

            try:
                component_usage = marketplace_models.ComponentUsage.objects.get(
                    resource=resource,
                    component=offering_component,
                    billing_period=month_start(date),
                    plan_period=plan_period,
                )
                component_usage.usage = max(usage, component_usage.usage)
                component_usage.save()
            except django_exceptions.ObjectDoesNotExist:
                marketplace_models.ComponentUsage.objects.create(
                    resource=resource,
                    component=offering_component,
                    usage=usage,
                    date=date,
                    billing_period=month_start(date),
                    plan_period=plan_period,
                )

        except marketplace_models.OfferingComponent.DoesNotExist:
            logger.warning('Skipping node usage synchronization because this '
                           'marketplace.OfferingComponent does not exist.'
                           'Cluster ID: %s', cluster.id)
        except marketplace_models.ResourcePlanPeriod.DoesNotExist:
            logger.warning('Skipping node usage synchronization because this '
                           'marketplace.ResourcePlanPeriod does not exist.'
                           'Cluster ID: %s', cluster.id)


def component_usage_register(component_usage):
    offering_component = component_usage.component

    plan_period = component_usage.plan_period
    if not plan_period:
        logger.warning('Skipping processing of component usage with ID %s because '
                       'plan period is not defined.', component_usage.id)
        return
    plan = plan_period.plan

    try:
        plan_component = plan.components.get(component=offering_component)
        item = invoice_models.InvoiceItem.objects.get(scope=component_usage.resource.scope,
                                                      details__plan_period_id=plan_period.id,
                                                      details__plan_component_id=plan_component.id,
                                                      invoice__year=component_usage.billing_period.year,
                                                      invoice__month=component_usage.billing_period.month)
        item.quantity = component_usage.usage
        item.unit_price = plan_component.price
        item.save()
    except invoice_models.InvoiceItem.DoesNotExist:
        cluster = component_usage.resource.scope
        customer = cluster.customer
        invoice, created = registrators.RegistrationManager.get_or_create_invoice(customer, component_usage.date)
        details = {
            'cluster_id': cluster.id,
            'offering_id': offering_component.offering.id,
            'plan_period_id': plan_period.id,
            'plan_component_id': plan_component.id
        }
        month_start = core_utils.month_start(component_usage.date)
        month_end = core_utils.month_end(component_usage.date)

        start = month_start if not component_usage.plan_period.start else \
            max(component_usage.plan_period.start, month_start)
        end = month_end if not component_usage.plan_period.end else \
            min(component_usage.plan_period.end, month_end)

        invoice_models.InvoiceItem.objects.create(
            content_type=ContentType.objects.get_for_model(cluster),
            object_id=cluster.id,
            project=cluster.project,
            invoice=invoice,
            start=start,
            end=end,
            details=details,
            unit_price=plan_component.price,
            quantity=component_usage.usage,
            unit=common_mixins.UnitPriceMixin.Units.QUANTITY,
            product_code=offering_component.product_code or plan.product_code,
            article_code=offering_component.article_code or plan.article_code,
            name=component_usage.resource.name + '(%s)' % offering_component.offering.name,
        )

    except marketplace_models.PlanComponent.DoesNotExist:
        logger.warning('Plan component for usage component %s is not found.', component_usage.id)
    except invoice_models.InvoiceItem.MultipleObjectsReturned:
        logger.warning('Skipping the invoice item unit price update '
                       'because multiple GenericInvoiceItem objects found. Scope: %s %s, date: %s.',
                       component_usage.resource.content_type,
                       component_usage.resource.object_id,
                       component_usage.date)



""" Defines how to optimize OpenStack packages """
from rest_framework import serializers as rf_serializers

from waldur_cost_planning import optimizers, register, serializers
from waldur_openstack.openstack import apps as openstack_apps

from waldur_mastermind.packages.models import PackageTemplate, PackageComponent
from waldur_mastermind.packages.serializers import PackageTemplateSerializer


OptimizedOpenStack = optimizers.namedtuple_with_defaults(
    'OptimizedOpenStack',
    field_names=optimizers.OptimizedService._fields + ('package_template',),
    default_values=optimizers.OptimizedService._defaults,
)


class OpenStackOptimizer(optimizers.Optimizer):
    """ Find the cheapest package template for OpenStack service """

    def optimize(self, deployment_plan, service):
        requirements = deployment_plan.get_requirements()
        # Step 1. Find suitable templates.
        templates = PackageTemplate.objects.filter(
            service_settings=service.settings).prefetch_related('components')
        suitable_templates = []
        for template in templates:
            components = {c.type: c.amount for c in template.components.all()}
            is_suitable = (
                components[PackageComponent.Types.RAM] >= requirements['ram'] and
                components[PackageComponent.Types.STORAGE] >= requirements['storage'] and
                components[PackageComponent.Types.CORES] >= requirements['cores']
            )
            if is_suitable:
                suitable_templates.append(template)
        # Step 2. Find the cheapest template.
        if not suitable_templates:
            raise optimizers.OptimizationError('There is no package template that can hold the deployment plan.')
        cheapest_template = min(suitable_templates, key=lambda t: t.price)
        return OptimizedOpenStack(service, cheapest_template.price, '', cheapest_template)


register.Register.register_optimizer(openstack_apps.OpenStackConfig.service_name, OpenStackOptimizer)


class OptimizedOpenStackSerializer(serializers.OptimizedServiceSerializer):
    service = rf_serializers.HyperlinkedRelatedField(
        view_name='openstack-detail',
        lookup_field='uuid',
        read_only=True,
    )

    def get_fields(self):
        fields = super(OptimizedOpenStackSerializer, self).get_fields()
        fields['package_template'] = PackageTemplateSerializer(read_only=True)
        return fields


register.Register.register_serializer(openstack_apps.OpenStackConfig.service_name, OptimizedOpenStackSerializer)



# Generated by Django 2.2.7 on 2019-11-18 09:48

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('waldur_rancher', '0004_node_backend_id'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='node',
            unique_together={('cluster', 'name'), ('content_type', 'object_id')},
        ),
    ]



from waldur_core.structure import serializers as structure_serializers

from . import models


class BaseVolumeTypeSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.BaseVolumeType
        fields = ('url', 'uuid', 'name', 'description', 'settings')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'settings': {
                'lookup_field': 'uuid',
                'view_name': 'servicesettings-detail',
            },
        }



from celery import shared_task

from waldur_core.core import utils as core_utils

from . import utils


@shared_task(name='waldur_mastermind.booking.send_notifications_about_upcoming_bookings')
def send_notifications_about_upcoming_bookings():
    for info in utils.get_info_about_upcoming_bookings():
        emails = [info['user'].email]
        core_utils.broadcast_mail('booking', 'notification', info, emails)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-04 20:54
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0002_immutable_default_json'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='defaultpricelistitem',
            name='metadata',
        ),
    ]



from django.apps import AppConfig


class MarketplaceRancherConfig(AppConfig):
    name = 'waldur_mastermind.marketplace_rancher'
    verbose_name = 'Marketplace Rancher'

    def ready(self):
        from waldur_core.structure import signals as structure_signals
        from waldur_mastermind.marketplace import handlers as marketplace_handlers
        from waldur_mastermind.marketplace import models as marketplace_models
        from waldur_mastermind.marketplace.plugins import manager, Component
        from waldur_rancher.apps import RancherConfig
        from waldur_rancher import models as rancher_models

        from . import handlers, PLUGIN_NAME, processors

        USAGE = marketplace_models.OfferingComponent.BillingTypes.USAGE
        manager.register(offering_type=PLUGIN_NAME,
                         create_resource_processor=processors.RancherCreateProcessor,
                         delete_resource_processor=processors.RancherDeleteProcessor,
                         components=(
                             Component(type='node', name='K8S node', measured_unit='', billing_type=USAGE),
                         ),
                         service_type=RancherConfig.service_name,
                         resource_model=rancher_models.Cluster)

        marketplace_handlers.connect_resource_metadata_handlers(rancher_models.Cluster)
        marketplace_handlers.connect_resource_handlers(rancher_models.Cluster)

        structure_signals.resource_imported.connect(
            handlers.create_marketplace_resource_for_imported_cluster,
            sender=rancher_models.Cluster,
            dispatch_uid='waldur_mastermind.marketpace_rancher.'
                         'create_resource_for_imported_cluster',
        )



import json
import logging
import os
import subprocess  # nosec

from django.conf import settings

from waldur_ansible.common.exceptions import AnsibleBackendError
from waldur_core.core.views import RefreshTokenMixin

logger = logging.getLogger(__name__)


class AnsiblePlaybookBackend:
    def __init__(self, playbook):
        self.playbook = playbook

    def _get_command(self, job, check_mode):
        playbook_path = self.playbook.get_playbook_path()
        if not os.path.exists(playbook_path):
            raise AnsibleBackendError('Playbook %s does not exist.' % playbook_path)

        command = [settings.WALDUR_ANSIBLE_COMMON.get('PLAYBOOK_EXECUTION_COMMAND', 'ansible-playbook')]
        if settings.WALDUR_ANSIBLE_COMMON.get('PLAYBOOK_ARGUMENTS'):
            command.extend(settings.WALDUR_ANSIBLE_COMMON.get('PLAYBOOK_ARGUMENTS'))

        if check_mode:
            command.append('--check')

        extra_vars = job.arguments.copy()
        extra_vars.update(self._get_extra_vars(job))
        # XXX: Passing arguments in following way is supported in Ansible>=1.2
        command.extend(['--extra-vars', json.dumps(extra_vars)])

        command.extend(['--ssh-common-args', '-o UserKnownHostsFile=/dev/null'])
        return command + [playbook_path]

    def _get_extra_vars(self, job):
        return dict(
            api_url=settings.WALDUR_ANSIBLE_COMMON['API_URL'],
            access_token=RefreshTokenMixin().refresh_token(job.user).key,
            project_uuid=job.service_project_link.project.uuid.hex,
            provider_uuid=job.service_project_link.service.uuid.hex,
            private_key_path=settings.WALDUR_ANSIBLE_COMMON['PRIVATE_KEY_PATH'],
            public_key_uuid=settings.WALDUR_ANSIBLE_COMMON['PUBLIC_KEY_UUID'],
            user_key_uuid=job.ssh_public_key.uuid.hex,
            subnet_uuid=job.subnet.uuid.hex,
            tags=[job.get_tag()],
        )

    def run_job(self, job, check_mode=False):
        command = self._get_command(job, check_mode)
        command_str = ' '.join(command)

        logger.debug('Executing command "%s".', command_str)
        env = dict(
            os.environ,
            ANSIBLE_LIBRARY=settings.WALDUR_ANSIBLE_COMMON['ANSIBLE_LIBRARY'],
            ANSIBLE_HOST_KEY_CHECKING='False',
        )
        try:
            output = subprocess.check_output(command,  # nosec
                                             stderr=subprocess.STDOUT,
                                             env=env,
                                             encoding='utf-8')
        except subprocess.CalledProcessError as e:
            logger.info('Failed to execute command "%s".', command_str)
            job.output = e.output
            job.save(update_fields=['output'])
            raise AnsibleBackendError(e)
        else:
            logger.info('Command "%s" was successfully executed.', command_str)
            job.output = output
            job.save(update_fields=['output'])

    def decode_output(self, output):
        items = []
        for line in output.splitlines():
            if 'WALDUR_CHECK_MODE' not in line:
                continue
            parts = line.split(' => ')
            if len(parts) != 2:
                continue
            try:
                payload = json.loads(parts[1])
            except TypeError:
                continue
            if 'instance' in payload:
                payload = payload['instance']
            if 'WALDUR_CHECK_MODE' in payload:
                del payload['WALDUR_CHECK_MODE']
            if payload:
                items.append(payload)
        return items



from . import views


def register_in(router):
    router.register(r'digitalocean', views.DigitalOceanServiceViewSet, basename='digitalocean')
    router.register(r'digitalocean-images', views.ImageViewSet, basename='digitalocean-image')
    router.register(r'digitalocean-regions', views.RegionViewSet, basename='digitalocean-region')
    router.register(r'digitalocean-sizes', views.SizeViewSet, basename='digitalocean-size')
    router.register(r'digitalocean-droplets', views.DropletViewSet, basename='digitalocean-droplet')
    router.register(r'digitalocean-service-project-link',
                    views.DigitalOceanServiceProjectLinkViewSet, basename='digitalocean-spl')



from waldur_core.core import WaldurExtension


class AuthSocialExtension(WaldurExtension):

    class Settings:
        # wiki: https://opennode.atlassian.net/wiki/display/WD/AuthSocial+plugin+configuration
        WALDUR_AUTH_SOCIAL = {
            'GOOGLE_SECRET': '',
            'GOOGLE_CLIENT_ID': '',

            'FACEBOOK_SECRET': '',
            'FACEBOOK_CLIENT_ID': '',

            'SMARTIDEE_SECRET': '',
            'SMARTIDEE_CLIENT_ID': '',

            'TARA_SECRET': '',
            'TARA_CLIENT_ID': '',
            'TARA_SANDBOX': True,
            'TARA_LABEL': 'Riigi Autentimisteenus',

            'USER_ACTIVATION_URL_TEMPLATE': 'http://example.com/#/activate/{user_uuid}/{token}/',
        }

    @staticmethod
    def get_public_settings():
        return [
            'GOOGLE_CLIENT_ID',
            'FACEBOOK_CLIENT_ID',
            'SMARTIDEE_CLIENT_ID',
            'TARA_CLIENT_ID',
            'TARA_SANDBOX',
            'TARA_LABEL',
        ]

    @staticmethod
    def django_app():
        return 'waldur_auth_social'

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns



from celery import chain

from waldur_core.core import executors as core_executors
from waldur_core.core import tasks as core_tasks


class VolumePullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_volume, 'pull_volume',
            state_transition='begin_updating')


class VolumeCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return chain(
            core_tasks.BackendMethodTask().si(
                serialized_volume,
                'create_volume',
                state_transition='begin_creating'
            ),
            core_tasks.PollRuntimeStateTask().si(
                serialized_volume,
                backend_pull_method='pull_volume_runtime_state',
                success_state='available',
                erred_state='error',
            ).set(countdown=30)
        )


class VolumeDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        if volume.backend_id:
            return chain(
                core_tasks.BackendMethodTask().si(
                    serialized_volume, 'delete_volume', state_transition='begin_deleting'),
                core_tasks.PollBackendCheckTask().si(serialized_volume, 'is_volume_deleted'),
            )
        else:
            return core_tasks.StateTransitionTask().si(serialized_volume, state_transition='begin_deleting')


class InstancePullExecutor(core_executors.ActionExecutor):
    action = 'Pull'

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_volume, 'pull_instance',
            state_transition='begin_updating')


class InstanceCreateExecutor(core_executors.CreateExecutor):

    @classmethod
    def get_task_signature(cls, volume, serialized_volume, **kwargs):
        return core_tasks.BackendMethodTask().si(
            serialized_volume,
            'create_instance',
            state_transition='begin_creating'
        )


class InstanceDeleteExecutor(core_executors.DeleteExecutor):

    @classmethod
    def get_task_signature(cls, instance, serialized_instance, **kwargs):
        if instance.backend_id:
            return cls.get_delete_instance_tasks(instance, serialized_instance)
        else:
            return core_tasks.StateTransitionTask().si(serialized_instance, state_transition='begin_deleting')

    @classmethod
    def get_delete_instance_tasks(cls, instance, serialized_instance):
        _tasks = [
            core_tasks.BackendMethodTask().si(
                serialized_instance,
                'delete_instance',
                state_transition='begin_deleting'
            ),
            core_tasks.PollBackendCheckTask().si(
                serialized_instance,
                'is_instance_deleted'
            ),
            core_tasks.IndependentBackendMethodTask().si(
                serialized_instance,
                'pull_networks'
            ),
        ]

        if instance.floating_ip:
            _tasks.append(core_tasks.IndependentBackendMethodTask().si(
                serialized_instance,
                'pull_floating_ips'
            ))

        return chain(_tasks)



from django.core.exceptions import ObjectDoesNotExist, MultipleObjectsReturned
from rest_framework import serializers
from django.utils.translation import ugettext_lazy as _

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers
from waldur_vmware.utils import is_basic_mode

from . import constants, models


def get_int_or_none(options, key):
    value = options.get(key)
    if not value:
        return value

    if isinstance(value, str):
        try:
            return int(value)
        except ValueError:
            return

    return value


class OptionsSerializer(serializers.Serializer):
    default_cluster_label = serializers.CharField(required=False)
    max_cpu = core_serializers.UnicodeIntegerField(min_value=1, required=False)
    max_cores_per_socket = core_serializers.UnicodeIntegerField(min_value=1, required=False)
    max_ram = core_serializers.UnicodeIntegerField(min_value=1, required=False)
    max_disk = core_serializers.UnicodeIntegerField(min_value=1, required=False)
    max_disk_total = core_serializers.UnicodeIntegerField(min_value=1, required=False)


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': _('VMware auth URL'),
        'username': _('VMware user username'),
        'password': _('VMware user password'),
    }

    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'default_cluster_label': _('Label of VMware cluster that will be used for virtual machines provisioning'),
        'max_cpu': _('Maximum vCPU for each VM'),
        'max_cores_per_socket': _('Maximum number of cores per socket for each VM'),
        'max_ram': _('Maximum RAM for each VM, MiB'),
        'max_disk': _('Maximum capacity for each disk, MiB'),
        'max_disk_total': _('Maximum total size of the disk space per VM, MiB'),
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.VMwareService
        required_fields = ('backend_url', 'username', 'password', 'default_cluster_label')
        options_serializer = OptionsSerializer


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.VMwareServiceProjectLink
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'vmware-detail'},
        }


class LimitSerializer(serializers.Serializer):
    def to_representation(self, service_settings):
        fields = (
            'max_cpu',
            'max_cores_per_socket',
            'max_ram',
            'max_disk',
            'max_disk_total',
        )
        result = dict()
        for field in fields:
            result[field] = get_int_or_none(service_settings.options, field)
        return result


class NestedPortSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Port
        fields = ('url', 'uuid', 'name', 'mac_address', 'network')
        read_only_fields = ('mac_address',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'vmware-port-detail'},
            'network': {'lookup_field': 'uuid', 'view_name': 'vmware-network-detail'},
        }


class NestedDiskSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = models.Disk
        fields = ('url', 'uuid', 'size')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'vmware-disk-detail'},
        }


class NestedNetworkSerializer(core_serializers.AugmentedSerializerMixin,
                              core_serializers.HyperlinkedRelatedModelSerializer):
    class Meta:
        model = models.Network
        fields = ('uuid', 'url', 'name', 'type')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class VirtualMachineSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='vmware-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='vmware-spl-detail',
        queryset=models.VMwareServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    guest_os = serializers.ChoiceField(
        choices=list(constants.GUEST_OS_CHOICES.items()),
        required=False,
        allow_null=True,
    )

    guest_os_name = serializers.SerializerMethodField()

    disks = NestedDiskSerializer(many=True, read_only=True)

    ports = NestedPortSerializer(many=True, read_only=True, source='port_set')

    template = serializers.HyperlinkedRelatedField(
        view_name='vmware-template-detail',
        lookup_field='uuid',
        queryset=models.Template.objects.all(),
        allow_null=True,
        required=False,
        write_only=True,
    )

    template_name = serializers.ReadOnlyField(source='template.name')

    cluster = serializers.HyperlinkedRelatedField(
        view_name='vmware-cluster-detail',
        lookup_field='uuid',
        queryset=models.Cluster.objects.all(),
        allow_null=True,
        required=False,
    )

    cluster_name = serializers.ReadOnlyField(source='cluster.name')

    datastore = serializers.HyperlinkedRelatedField(
        view_name='vmware-datastore-detail',
        lookup_field='uuid',
        queryset=models.Datastore.objects.all(),
        allow_null=True,
        required=False,
    )

    datastore_name = serializers.ReadOnlyField(source='datastore.name')

    folder = serializers.HyperlinkedRelatedField(
        view_name='vmware-folder-detail',
        lookup_field='uuid',
        queryset=models.Folder.objects.all(),
        allow_null=True,
        required=False,
    )

    folder_name = serializers.ReadOnlyField(source='folder.name')

    networks = NestedNetworkSerializer(queryset=models.Network.objects.all(),
                                       many=True, required=False, write_only=True)

    runtime_state = serializers.SerializerMethodField()

    tools_state = serializers.ReadOnlyField(source='get_tools_state_display')

    def get_runtime_state(self, vm):
        return dict(models.VirtualMachine.RuntimeStates.CHOICES).get(vm.runtime_state)

    def get_guest_os_name(self, vm):
        return constants.GUEST_OS_CHOICES.get(vm.guest_os)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.VirtualMachine
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'guest_os', 'guest_os_name', 'cores', 'cores_per_socket', 'ram', 'disk', 'disks',
            'runtime_state', 'template', 'cluster', 'networks', 'datastore', 'folder',
            'template_name', 'cluster_name', 'datastore_name', 'folder_name', 'ports',
            'guest_power_state', 'tools_state', 'tools_installed',
        )
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'guest_os', 'template', 'cluster', 'networks', 'datastore', 'folder', 'ports',
            'name',
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'disk', 'runtime_state', 'guest_power_state', 'tools_installed',
        )
        extra_kwargs = dict(
            cores={'required': False},
            cores_per_socket={'required': False},
            ram={'required': False},
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def get_fields(self):
        """
        When basic mode is activated, user is not allowed
        to select placement attributes for the new virtual machine.
        """
        fields = super(VirtualMachineSerializer, self).get_fields()

        if 'ram' in fields:
            fields['ram'].factor = 1024
            fields['ram'].units = 'GB'
            fields['ram'].min_value = 1024

        if 'disk' in fields:
            fields['disk'].factor = 1024
            fields['disk'].units = 'GB'

        if 'cores' in fields:
            fields['cores'].min_value = 1

        if 'cores_per_socket' in fields:
            fields['cores_per_socket'].min_value = 1

        if isinstance(self.instance, models.VirtualMachine):
            spl = self.instance.service_project_link
            options = spl.service.settings.options

            if 'cores' in fields:
                fields['cores'].max_value = options.get('max_cpu')

            if 'ram' in fields and 'max_ram' in options:
                fields['ram'].max_value = options.get('max_ram')

            if 'cores_per_socket' in fields and 'max_cores_per_socket' in options:
                fields['cores_per_socket'].max_value = options.get('max_cores_per_socket')

            if 'disk' in fields and 'max_disk' in options:
                fields['disk'].max_value = get_int_or_none(options, 'max_disk')

            if 'disk' in fields and 'max_disk_total' in options:
                if fields['disk'].max_value:
                    fields['disk'].max_value = min(
                        get_int_or_none(options, 'max_disk_total'),
                        get_int_or_none(options, 'max_disk')
                    )
                else:
                    fields['disk'].max_value = get_int_or_none(options, 'max_disk_total')

        if not is_basic_mode():
            return fields

        try:
            method = self.context['view'].request.method
        except (KeyError, AttributeError):
            return fields

        if method == 'POST':
            read_only_fields = 'cluster', 'networks', 'datastore', 'folder'
            for field in read_only_fields:
                fields[field].read_only = True

        return fields

    def _validate_attributes(self, attrs):
        template_attrs = {'cores', 'cores_per_socket', 'ram', 'disk', 'guest_os'}
        template = attrs.get('template')
        missing_attributes = template_attrs - set(attrs.keys()) - {'disk'}
        if template:
            if attrs.get('guest_os'):
                raise serializers.ValidationError(
                    'It is not possible to customize guest OS when template is used.')
            for attr in template_attrs:
                old_value = attrs.get(attr)
                if not old_value:
                    attrs[attr] = getattr(template, attr)
        elif missing_attributes:
            attr_list = ', '.join(missing_attributes)
            raise serializers.ValidationError(
                'These fields are required when template is not used: %s.' % attr_list)
        return attrs

    def _validate_cpu(self, attrs, options):
        """
        Validate vCPU specification against service limits.
        """
        actual_cpu = attrs.get('cores')
        max_cpu = options.get('max_cpu')
        if actual_cpu and max_cpu and actual_cpu > max_cpu:
            raise serializers.ValidationError('Requested amount of CPU exceeds offering limit.')

        cores_per_socket = attrs.get('cores_per_socket')
        if cores_per_socket and actual_cpu % cores_per_socket != 0:
            raise serializers.ValidationError('Number of CPU cores should be multiple of cores per socket.')

        max_cores_per_socket = options.get('max_cores_per_socket')
        if cores_per_socket and max_cores_per_socket and cores_per_socket > max_cores_per_socket:
            raise serializers.ValidationError('Requested amount of cores per socket exceeds offering limit.')

    def _validate_ram(self, attrs, options):
        """
        Validate RAM specification against service limits.
        """
        actual_ram = attrs.get('ram')
        max_ram = options.get('max_ram')
        if actual_ram and actual_ram < 1024:
            raise serializers.ValidationError('Requested amount of RAM is too small.')

        if actual_ram and max_ram and actual_ram > max_ram:
            raise serializers.ValidationError('Requested amount of RAM exceeds offering limit.')

    def _validate_disk(self, attrs, options):
        """
        Validate storage specification against service limits.
        """
        template = attrs.get('template')
        max_disk = get_int_or_none(options, 'max_disk')
        actual_disk = template.disk if template else 0
        if actual_disk and max_disk and actual_disk > max_disk:
            raise serializers.ValidationError('Requested amount of disk exceeds offering limit.')

        max_disk_total = get_int_or_none(options, 'max_disk_total')
        if actual_disk and max_disk_total and actual_disk > max_disk_total:
            raise serializers.ValidationError('Requested amount of disk exceeds offering limit.')

    def _validate_limits(self, attrs):
        """
        Validate hardware specification against service limits.
        """
        if self.instance:
            spl = self.instance.service_project_link
        else:
            spl = attrs['service_project_link']

        options = spl.service.settings.options
        self._validate_cpu(attrs, options)
        self._validate_ram(attrs, options)
        self._validate_disk(attrs, options)
        return attrs

    def _validate_cluster(self, attrs):
        """
        If basic mode is activated, match cluster by customer and service.
        Otherwise use cluster provided by user if it is allowed.
        Finally, use default cluster from service settings.
        """
        spl = attrs['service_project_link']

        if is_basic_mode():
            customer = spl.project.customer
            try:
                cluster = models.Cluster.objects.filter(
                    settings=spl.service.settings,
                    customercluster__customer=customer).get()
            except ObjectDoesNotExist:
                return self._fallback_to_default_cluster(attrs)
            except MultipleObjectsReturned:
                raise serializers.ValidationError(
                    'There are multiple clusters assigned to the current customer.')
            else:
                attrs['cluster'] = cluster
            return attrs

        cluster = attrs.get('cluster')

        if cluster:
            if cluster.settings != spl.service.settings:
                raise serializers.ValidationError('This cluster is not available for this service.')

            if not cluster.customercluster_set.filter(customer=spl.project.customer).exists():
                raise serializers.ValidationError('This cluster is not available for this customer.')
        else:
            return self._fallback_to_default_cluster(attrs)
        return attrs

    def _fallback_to_default_cluster(self, attrs):
        spl = attrs['service_project_link']
        default_cluster_label = spl.service.settings.options.get('default_cluster_label')

        if not default_cluster_label:
            raise serializers.ValidationError('Default cluster is not defined for this service.')
        try:
            attrs['cluster'] = models.Cluster.objects.filter(
                settings=spl.service.settings,
                name=default_cluster_label).get()
            return attrs
        except models.Cluster.DoesNotExist:
            raise serializers.ValidationError('Default cluster is not defined for this service.')

    def _validate_folder(self, attrs):
        """
        If basic mode is activated, match folder by customer and service.
        Otherwise use folder provided by user if it is allowed.
        """
        spl = attrs['service_project_link']

        if is_basic_mode():
            customer = spl.project.customer
            try:
                folder = models.Folder.objects.filter(
                    settings=spl.service.settings,
                    customerfolder__customer=customer).get()
            except ObjectDoesNotExist:
                raise serializers.ValidationError(
                    'There is no folder assigned to the current customer.')
            except MultipleObjectsReturned:
                raise serializers.ValidationError(
                    'There are multiple folders assigned to the current customer.')
            else:
                attrs['folder'] = folder
            return attrs

        folder = attrs.get('folder')

        if folder:
            if folder.settings != spl.service.settings:
                raise serializers.ValidationError('This folder is not available for this service.')

            if not folder.customerfolder_set.filter(customer=spl.project.customer).exists():
                raise serializers.ValidationError('This folder is not available for this customer.')
        return attrs

    def _validate_datastore(self, attrs):
        """
        If basic mode is activated, match datastore by customer and service.
        Otherwise use datastore provided by user if it is valid with respect to its size.
        """
        spl = attrs['service_project_link']
        template = attrs.get('template')

        if is_basic_mode():
            customer = spl.project.customer
            datastore = models.Datastore.objects.filter(
                settings=spl.service.settings,
                customerdatastore__customer=customer
            ).order_by('-free_space').first()
            if not datastore:
                raise serializers.ValidationError(
                    'There is no datastore assigned to the current customer.')
            elif template and template.disk > datastore.free_space:
                raise serializers.ValidationError(
                    'There is no datastore with enough free space available for current customer.')
            else:
                attrs['datastore'] = datastore
            return attrs

        datastore = attrs.get('datastore')

        if datastore:
            if datastore.settings != spl.service.settings:
                raise serializers.ValidationError('This datastore is not available for this service.')

            if not datastore.customerdatastore_set.filter(customer=spl.project.customer).exists():
                raise serializers.ValidationError('This datastore is not available for this customer.')

            if template and template.disk > datastore.free_space:
                raise serializers.ValidationError(
                    'There is no datastore with enough free space available for current customer.')
        return attrs

    def _validate_networks(self, attrs):
        """
        If basic mode is activated, match network by customer and service.
        Otherwise use networks provided by user if it is valid with respect to its size.
        """
        spl = attrs['service_project_link']

        if is_basic_mode():
            customer = spl.project.customer
            try:
                network = models.Network.objects.filter(
                    settings=spl.service.settings,
                    customernetwork__customer=customer).get()
            except ObjectDoesNotExist:
                raise serializers.ValidationError(
                    'There is no network assigned to the current customer.')
            except MultipleObjectsReturned:
                raise serializers.ValidationError(
                    'There are multiple networks assigned to the current customer.')
            else:
                attrs['networks'] = [network]
            return attrs

        networks = attrs.get('networks', [])

        for network in networks:
            if network.settings != spl.service.settings:
                raise serializers.ValidationError('This network is not available for this service.')

            if not network.customernetwork_set.filter(customer=spl.project.customer).exists():
                raise serializers.ValidationError('This network is not available for this customer.')

        return attrs

    validators_pipeline = [
        _validate_attributes,
        _validate_cluster,
        _validate_limits,
        _validate_datastore,
        _validate_folder,
        _validate_networks,
    ]

    def validate(self, attrs):
        attrs = super(VirtualMachineSerializer, self).validate(attrs)

        if self.instance:
            self._validate_limits(attrs)
        else:
            for validator in self.validators_pipeline:
                attrs = validator(self, attrs)

        return attrs

    def create(self, validated_data):
        networks = validated_data.pop('networks', [])
        vm = super(VirtualMachineSerializer, self).create(validated_data)
        vm.networks.add(*networks)
        return vm


class PortSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='vmware-detail',
        read_only=True,
        lookup_field='uuid',
    )

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='vmware-spl-detail',
        read_only=True,
    )

    service_settings = serializers.HyperlinkedRelatedField(
        view_name='servicesettings-detail',
        lookup_field='uuid',
        read_only=True,
    )

    project = serializers.HyperlinkedRelatedField(
        view_name='project-detail',
        lookup_field='uuid',
        read_only=True,
    )

    vm_name = serializers.ReadOnlyField(source='vm.name')
    vm_uuid = serializers.ReadOnlyField(source='vm.uuid')
    network_name = serializers.ReadOnlyField(source='network.name')

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Port
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'mac_address', 'vm', 'vm_uuid', 'vm_name', 'network', 'network_name',
        )
        # Virtual Ethernet adapter name is generated automatically by VMware itself,
        # therefore it's not editable by user
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'name', 'vm', 'mac_address',
        )
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'network',
        )
        extra_kwargs = dict(
            vm={
                'view_name': 'vmware-virtual-machine-detail',
                'lookup_field': 'uuid',
            },
            network={
                'view_name': 'vmware-network-detail',
                'lookup_field': 'uuid',
            },
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def get_fields(self):
        # Skip metadata tweak on update because network is protected from update
        fields = super(PortSerializer, self).get_fields()
        if 'network' in fields and self.instance:
            network_field = fields['network']
            network_field.display_name_field = 'name'
            network_field.query_params = {
                'customer_pair_uuid': self.instance.customer.uuid.hex,
                'settings_uuid': self.instance.service_settings.uuid.hex,
            }
        return fields

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        vm = self.context['view'].get_object()
        attrs['vm'] = vm
        attrs['service_project_link'] = vm.service_project_link

        if not models.CustomerNetworkPair.objects.filter(
            customer=vm.customer,
            network=attrs['network'],
        ).exists():
            raise serializers.ValidationError('This network is not available for this customer.')

        return super(PortSerializer, self).validate(attrs)

    def create(self, validated_data):
        # Virtual Adapter is updated with actual name when pulling is performed
        validated_data['name'] = 'New virtual Adapter'
        return super(PortSerializer, self).create(validated_data)


class DiskSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='vmware-detail',
        read_only=True,
        lookup_field='uuid',
    )

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='vmware-spl-detail',
        read_only=True,
    )

    service_settings = serializers.HyperlinkedRelatedField(
        view_name='servicesettings-detail',
        lookup_field='uuid',
        read_only=True,
    )

    project = serializers.HyperlinkedRelatedField(
        view_name='project-detail',
        lookup_field='uuid',
        read_only=True,
    )

    vm_uuid = serializers.ReadOnlyField(source='vm.uuid')
    vm_name = serializers.ReadOnlyField(source='vm.name')

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Disk
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'size', 'vm', 'vm_uuid', 'vm_name'
        )
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'size', 'name'
        )
        # Virtual disk name is generated automatically by VMware itself,
        # therefore it's not editable by user
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'name', 'vm',
        )
        extra_kwargs = dict(
            vm={
                'view_name': 'vmware-virtual-machine-detail',
                'lookup_field': 'uuid',
            },
            **structure_serializers.BaseResourceSerializer.Meta.extra_kwargs
        )

    def get_fields(self):
        fields = super(DiskSerializer, self).get_fields()
        fields['size'].factor = 1024
        fields['size'].units = 'GB'
        fields['size'].min_value = 1024

        if isinstance(self.instance, models.VirtualMachine):
            options = self.instance.service_settings.options

            max_disk = get_int_or_none(options, 'max_disk')
            if max_disk:
                fields['size'].max_value = max_disk

            max_disk_total = get_int_or_none(options, 'max_disk_total')
            if max_disk_total:
                remaining_quota = max_disk_total - self.instance.total_disk
                if fields['size'].max_value:
                    fields['size'].max_value = min(max_disk, remaining_quota)
                else:
                    fields['size'].max_value = remaining_quota
        return fields

    def _validate_size(self, vm, attrs):
        options = vm.service_project_link.service.settings.options

        actual_disk = attrs.get('size')
        if actual_disk < 1024:
            raise serializers.ValidationError('Requested amount of disk is too small.')

        max_disk = get_int_or_none(options, 'max_disk')
        if actual_disk and max_disk and actual_disk > max_disk:
            raise serializers.ValidationError('Requested amount of disk exceeds offering limit.')

        max_disk_total = get_int_or_none(options, 'max_disk_total')
        if actual_disk and max_disk_total:
            remaining_quota = max_disk_total - vm.total_disk
            if actual_disk > remaining_quota:
                raise serializers.ValidationError('Requested amount of disk exceeds offering limit.')

    def create(self, validated_data):
        # Virtual disk is updated with actual name when pulling is performed
        validated_data['name'] = 'New disk'
        return super(DiskSerializer, self).create(validated_data)

    def validate(self, attrs):
        # Skip validation on update
        if self.instance:
            return attrs

        vm = self.context['view'].get_object()
        self._validate_size(vm, attrs)
        attrs['vm'] = vm
        attrs['service_project_link'] = vm.service_project_link
        return super(DiskSerializer, self).validate(attrs)


class DiskExtendSerializer(serializers.ModelSerializer):
    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Disk
        fields = ('size',)

    def get_fields(self):
        fields = super(DiskExtendSerializer, self).get_fields()
        fields['size'].factor = 1024
        fields['size'].units = 'GB'

        if isinstance(self.instance, models.Disk):
            fields['size'].min_value = self.instance.size + 1024
            options = self.instance.service_settings.options
            max_disk = get_int_or_none(options, 'max_disk')
            if max_disk:
                fields['size'].max_value = max_disk

            max_disk_total = get_int_or_none(options, 'max_disk_total')
            if max_disk_total:
                remaining_quota = max_disk_total - self.instance.vm.total_disk + self.instance.size
                if max_disk:
                    fields['size'].max_value = min(max_disk, remaining_quota)
                else:
                    fields['size'].max_value = remaining_quota

        return fields

    def validate_size(self, value):
        if value <= self.instance.size:
            raise serializers.ValidationError(
                _('Disk size should be greater than %s') % self.instance.size)

        options = self.instance.service_settings.options
        max_disk = get_int_or_none(options, 'max_disk')
        if max_disk and value > max_disk:
            raise serializers.ValidationError('Requested amount of disk exceeds offering limit.')

        max_disk_total = get_int_or_none(options, 'max_disk_total')
        if max_disk_total:
            remaining_quota = max_disk_total - self.instance.vm.total_disk + self.instance.size
            if value > remaining_quota:
                raise serializers.ValidationError('Requested amount of disk exceeds offering limit.')

        return value


class TemplateSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Template
        fields = (
            'url', 'uuid', 'name', 'description', 'created', 'modified',
            'guest_os', 'guest_os_name', 'cores', 'cores_per_socket', 'ram', 'disk',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }

    guest_os_name = serializers.SerializerMethodField()

    def get_guest_os_name(self, template):
        return constants.GUEST_OS_CHOICES.get(template.guest_os)


class ClusterSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Cluster
        fields = (
            'url', 'uuid', 'name',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class NetworkSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Network
        fields = (
            'url', 'uuid', 'name', 'type',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class DatastoreSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Datastore
        fields = (
            'url', 'uuid', 'name', 'type', 'capacity', 'free_space'
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class FolderSerializer(structure_serializers.BasePropertySerializer):
    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Folder
        fields = (
            'url', 'uuid', 'name',
        )
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }



from waldur_core.core import WaldurExtension


class MarketplaceChecklistExtension(WaldurExtension):

    @staticmethod
    def django_app():
        return 'waldur_mastermind.marketplace_checklist'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def django_urls():
        from .urls import urlpatterns
        return urlpatterns



import logging

from celery import shared_task
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import transaction
from django.db.models import Sum, Q
from django.utils import timezone
from rest_framework import status

from waldur_core.core import utils as core_utils
from waldur_core.structure import models as structure_models
from waldur_mastermind.common.utils import create_request
from waldur_mastermind.invoices import utils as invoice_utils
from waldur_mastermind.marketplace.utils import process_order_item

from . import utils, models, views, exceptions

logger = logging.getLogger(__name__)


def approve_order(order, user):
    order.approve()
    order.approved_by = user
    order.approved_at = timezone.now()
    order.save()

    serialized_order = core_utils.serialize_instance(order)
    serialized_user = core_utils.serialize_instance(user)
    transaction.on_commit(lambda: process_order.delay(serialized_order, serialized_user))
    transaction.on_commit(lambda: create_order_pdf.delay(order.pk))


@shared_task(name='marketplace.process_order')
def process_order(serialized_order, serialized_user):
    order = core_utils.deserialize_instance(serialized_order)
    user = core_utils.deserialize_instance(serialized_user)
    for item in order.items.all():
        process_order_item(item, user)


@shared_task(name='marketplace.create_screenshot_thumbnail')
def create_screenshot_thumbnail(uuid):
    screenshot = models.Screenshot.objects.get(uuid=uuid)
    utils.create_screenshot_thumbnail(screenshot)


@shared_task(name='marketplace.notify_order_approvers')
def notify_order_approvers(uuid):
    order = models.Order.objects.get(uuid=uuid)
    users = order.get_approvers()
    emails = [u.email for u in users if u.email]
    link_template = settings.WALDUR_MARKETPLACE['ORDER_LINK_TEMPLATE']

    context = {
        'order_url': link_template.format(project_uuid=order.project.uuid),
        'order': order,
        'site_name': settings.WALDUR_CORE['SITE_NAME'],
    }

    core_utils.broadcast_mail('marketplace', 'notification_approval', context, emails)


@shared_task(name='marketplace.notify_about_resource_change')
def notify_about_resource_change(event_type, context, resource_uuid):
    resource = models.Resource.objects.get(uuid=resource_uuid)
    emails = resource.project.get_users().values_list('email', flat=True)
    core_utils.broadcast_mail('marketplace', event_type, context, emails)


@shared_task
def create_order_pdf(order_id):
    order = models.Order.objects.get(pk=order_id)
    utils.create_order_pdf(order)


@shared_task
def create_pdf_for_all():
    for order in models.Order.objects.all():
        utils.create_order_pdf(order)


def filter_aggregate_by_scope(queryset, scope):
    scope_path = None

    if isinstance(scope, structure_models.Project):
        scope_path = 'resource__project'

    if isinstance(scope, structure_models.Customer):
        scope_path = 'resource__project__customer'

    if scope_path:
        queryset = queryset.filter(**{scope_path: scope})

    return queryset


def aggregate_reported_usage(start, end, scope):
    queryset = models.ComponentUsage.objects \
        .filter(date__gte=start, date__lte=end) \
        .exclude(component__parent=None)

    queryset = filter_aggregate_by_scope(queryset, scope)

    queryset = queryset.values('component__parent_id').annotate(total=Sum('usage'))

    return {
        row['component__parent_id']: row['total']
        for row in queryset
    }


def aggregate_fixed_usage(start, end, scope):
    queryset = models.ResourcePlanPeriod.objects.filter(
        # Resource has been active during billing period
        Q(start__gte=start, end__lte=end) |
        # Resource is still active
        Q(end__isnull=True) |
        # Resource has been launched in previous billing period and stopped in current
        Q(end__gte=start, end__lte=end)
    )
    queryset = filter_aggregate_by_scope(queryset, scope)

    queryset = queryset.values('plan__components__component__parent_id') \
        .annotate(total=Sum('plan__components__amount'))

    return {
        row['plan__components__component__parent_id']: row['total']
        for row in queryset
    }


def calculate_usage_for_scope(start, end, scope):
    reported_usage = aggregate_reported_usage(start, end, scope)
    fixed_usage = aggregate_fixed_usage(start, end, scope)
    # It needs to cover a case when a key is None because OfferingComponent.parent can be None.
    fixed_usage.pop(None, None)
    components = set(reported_usage.keys()) | set(fixed_usage.keys())
    content_type = ContentType.objects.get_for_model(scope)

    for component_id in components:
        models.CategoryComponentUsage.objects.update_or_create(
            content_type=content_type,
            object_id=scope.id,
            component_id=component_id,
            date=start,
            defaults={
                'reported_usage': reported_usage.get(component_id),
                'fixed_usage': fixed_usage.get(component_id),
            }
        )


@shared_task(name='waldur_mastermind.marketplace.calculate_usage_for_current_month')
def calculate_usage_for_current_month():
    start = invoice_utils.get_current_month_start()
    end = invoice_utils.get_current_month_end()
    scopes = []

    for customer in structure_models.Customer.objects.all():
        scopes.append(customer)
        for project in customer.projects.all():
            scopes.append(project)

    for scope in scopes:
        calculate_usage_for_scope(start, end, scope)


@shared_task(name='waldur_mastermind.marketplace.send_notifications_about_usages')
def send_notifications_about_usages():
    for warning in utils.get_info_about_missing_usage_reports():
        customer = warning['customer']
        emails = [owner.email for owner in customer.get_owners()]
        warning['public_resources_url'] = utils.get_public_resources_url(customer)

        if customer.serviceprovider.enable_notifications and emails:
            core_utils.broadcast_mail('marketplace', 'notification_usages', warning, emails)


@shared_task(name='marketplace.terminate_resource')
def terminate_resource(serialized_resource, serialized_user):
    resource = core_utils.deserialize_instance(serialized_resource)
    user = core_utils.deserialize_instance(serialized_user)
    view = views.ResourceViewSet.as_view({'post': 'terminate'})
    response = create_request(view, user, {}, uuid=resource.uuid.hex)

    if response.status_code != status.HTTP_200_OK:
        raise exceptions.ResourceTerminateException(response.rendered_content)



from django.core.exceptions import ObjectDoesNotExist
from django.core.management.base import BaseCommand, CommandError
from waldur_core.structure import models as structure_models

from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace_support import utils


class Command(BaseCommand):
    help = 'Import existing support offerings as marketplace resources.'

    def add_arguments(self, parser):
        parser.add_argument('--category', dest='category_uuid', required=True,
                            help='Specify a category to create offerings.')
        parser.add_argument('--customer', dest='customer_uuid', required=True,
                            help='Specify a customer to create offerings.')

    def handle(self, category_uuid, customer_uuid, *args, **options):
        try:
            category = marketplace_models.Category.objects.get(uuid=category_uuid)
        except ObjectDoesNotExist:
            raise CommandError('A category is not found.')
        try:
            customer = structure_models.Customer.objects.get(uuid=customer_uuid)
        except ObjectDoesNotExist:
            raise CommandError('A customer is not found.')

        offerings_counter, plans_counter, resources_counter = utils.init_offerings_and_resources(category, customer)

        self.stdout.write(self.style.SUCCESS('%s offerings have been created.' % offerings_counter))
        self.stdout.write(self.style.SUCCESS('%s plans have been created.' % plans_counter))
        self.stdout.write(self.style.SUCCESS('%s resources have been created.' % resources_counter))



import logging

from waldur_core.structure.permissions import _get_project
from waldur_mastermind.common.utils import mb_to_gb
from waldur_mastermind.invoices import models as invoices_models
from waldur_mastermind.invoices.registrators import BaseRegistrator
from waldur_mastermind.marketplace import models as marketplace_models
from waldur_mastermind.marketplace import utils as marketplace_utils
from waldur_vmware import models as vmware_models

logger = logging.getLogger(__name__)


class VirtualMachineRegistrator(BaseRegistrator):

    def get_customer(self, source):
        return source.service_project_link.project.customer

    def get_sources(self, customer):
        return vmware_models.VirtualMachine.objects.filter(
            service_project_link__project__customer=customer
        ).exclude(backend_id=None).exclude(backend_id='').distinct()

    def _create_item(self, source, invoice, start, end):
        try:
            resource = marketplace_models.Resource.objects.get(scope=source)
            plan = resource.plan
            if not plan:
                logger.warning('Skipping VMware item invoice creation because '
                               'billing plan is not defined for resource. '
                               'Resource ID: %s', resource.id)
                return
        except marketplace_models.Resource.DoesNotExist:
            logger.warning('Skipping VMware item invoice creation because '
                           'marketplace resource is not available for VMware resource. '
                           'Resource ID: %s', source.id)
            return

        components_map = {
            plan_component.component.type: plan_component.price
            for plan_component in plan.components.all()
        }

        missing_components = {'cpu', 'ram', 'disk'} - set(components_map.keys())
        if missing_components:
            logger.warning('Skipping VMware item invoice creation because plan components are missing. '
                           'Plan ID: %s. Missing components: %s', plan.id, ', '.join(missing_components))
            return

        cores_price = components_map['cpu'] * source.cores
        ram_price = components_map['ram'] * mb_to_gb(source.ram)
        disk_price = components_map['disk'] * mb_to_gb(source.total_disk)
        total_price = cores_price + ram_price + disk_price

        start = invoices_models.adjust_invoice_items(
            invoice, source, start, total_price, plan.unit)

        details = self.get_details(source)
        item = invoices_models.InvoiceItem.objects.create(
            scope=source,
            project=_get_project(source),
            unit_price=total_price,
            unit=plan.unit,
            product_code=plan.product_code,
            article_code=plan.article_code,
            invoice=invoice,
            start=start,
            end=end,
            details=details,
        )
        self.init_details(item)

    def get_name(self, source):
        return '{name} ({cores} CPU, {ram} GB RAM, {disk} GB disk)'.format(
            name=source.name,
            cores=source.cores,
            ram=mb_to_gb(source.ram),
            disk=mb_to_gb(source.total_disk),
        )

    def get_details(self, source):
        details = {
            'cpu': source.cores,
            'ram': source.ram,
            'disk': source.total_disk,
        }
        service_provider_info = marketplace_utils.get_service_provider_info(source)
        details.update(service_provider_info)
        return details



from decimal import Decimal

from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import ugettext_lazy as _
from model_utils import FieldTracker

from waldur_core.core import models as core_models
from waldur_core.structure import models as structure_models
from waldur_openstack.openstack import models as openstack_models, apps as openstack_apps
from waldur_mastermind.common import mixins as common_mixins

from waldur_mastermind.common.utils import quantize_price


class PackageTemplate(core_models.UuidMixin,
                      core_models.NameMixin,
                      common_mixins.ProductCodeMixin,
                      core_models.UiDescribableMixin,):
    # We do not define permissions for PackageTemplate because we are planning
    # to use them with shared service settings only - it means that
    # PackageTemplates are visible for all users.
    service_settings = models.ForeignKey(on_delete=models.CASCADE, to=structure_models.ServiceSettings, related_name='+')
    archived = models.BooleanField(default=False, help_text=_('Forbids creation of new packages.'))
    tracker = FieldTracker()
    unit = models.CharField(default=common_mixins.UnitPriceMixin.Units.PER_DAY, max_length=30,
                            choices=common_mixins.UnitPriceMixin.Units.CHOICES)

    class Categories:
        SMALL = 'small'
        MEDIUM = 'medium'
        LARGE = 'large'
        TRIAL = 'trial'

        CHOICES = ((SMALL, _('Small')), (MEDIUM, _('Medium')), (LARGE, _('Large')), (TRIAL, _('Trial')))

    category = models.CharField(max_length=10, choices=Categories.CHOICES, default=Categories.SMALL)

    class Meta:
        verbose_name = _('VPC package template')
        verbose_name_plural = _('VPC package templates')

    @property
    def price(self):
        """
        Price for whole template for one day.
        :rtype: Decimal
        """
        if self.unit == common_mixins.UnitPriceMixin.Units.PER_DAY:
            price = self.components.aggregate(total=models.Sum(
                models.F('price') * models.F('amount'),
                output_field=models.DecimalField(max_digits=22, decimal_places=10)))['total'] or Decimal('0')
            return quantize_price(price)

    @property
    def monthly_price(self):
        """
        Return price for one month in case of monthly billing plan or price for 30 days otherwise.
        :rtype: Decimal
        """
        if self.unit == common_mixins.UnitPriceMixin.Units.PER_MONTH:
            price = self.components.aggregate(total=models.Sum(
                models.F('price') * models.F('amount'),
                output_field=models.DecimalField(max_digits=22, decimal_places=10)))['total'] or Decimal('0')
            return quantize_price(price)
        elif self.unit == common_mixins.UnitPriceMixin.Units.PER_DAY:
            return self.price * 30

    @staticmethod
    def get_required_component_types():
        return (PackageComponent.Types.RAM,
                PackageComponent.Types.CORES,
                PackageComponent.Types.STORAGE)

    @staticmethod
    def get_memory_types():
        return (PackageComponent.Types.RAM,
                PackageComponent.Types.STORAGE)

    def clean(self):
        openstack_type = openstack_apps.OpenStackConfig.service_name

        if not hasattr(self, 'service_settings'):
            raise ValidationError({'service_settings': _('Please select service settings.')})
        if not self.service_settings.shared:
            raise ValidationError({'service_settings': _('PackageTemplate can be created only for shared settings.')})
        if self.service_settings.type == openstack_type and not self.service_settings.options.get('is_admin', True):
            raise ValidationError({'service_settings': _('Service settings should support tenant creation.')})
        if 'external_network_id' not in self.service_settings.options:
            raise ValidationError(
                {'service_settings': _('external_network_id has to be defined for service settings.')})
        return self

    def is_read_only(self):
        return self.openstack_packages.exists()

    def __str__(self):
        return '%s | %s' % (self.name, self.service_settings.type)


class PackageComponent(models.Model):
    PRICE_MAX_DIGITS = 14
    PRICE_DECIMAL_PLACES = 10

    class Meta:
        unique_together = ('type', 'template')

    class Types:
        RAM = 'ram'
        CORES = 'cores'
        STORAGE = 'storage'

        CHOICES = ((RAM, 'RAM'), (CORES, 'Cores'), (STORAGE, 'Storage'))

    type = models.CharField(max_length=50, choices=Types.CHOICES)
    amount = models.PositiveIntegerField(default=0)
    price = models.DecimalField(default=0, max_digits=PRICE_MAX_DIGITS, decimal_places=PRICE_DECIMAL_PLACES,
                                validators=[MinValueValidator(Decimal('0'))],
                                verbose_name=_('Price per unit'))
    template = models.ForeignKey(on_delete=models.CASCADE, to=PackageTemplate, related_name='components')

    def __str__(self):
        return '%s | %s' % (self.type, self.template.name)

    @property
    def monthly_price(self):
        """
        Rounded price for 30-days.

        This price should not be used for calculations.
        Only to display price in human friendly way.
        """
        return round(self.price * 30 * self.amount, 2)


class OpenStackPackage(core_models.UuidMixin, models.Model):
    """ OpenStackPackage allows to create tenant and service_settings based on PackageTemplate """
    class Permissions:
        customer_path = 'tenant__service_project_link__project__customer'
        project_path = 'tenant__service_project_link__project'

    template = models.ForeignKey(PackageTemplate, related_name='openstack_packages',
                                 help_text=_('Tenant will be created based on this template.'),
                                 on_delete=models.PROTECT)
    tenant = models.OneToOneField(openstack_models.Tenant, related_name='+', on_delete=models.CASCADE)
    service_settings = models.ForeignKey(structure_models.ServiceSettings, related_name='+', null=True,
                                         on_delete=models.SET_NULL)

    def __str__(self):
        return 'Package "%s" for tenant %s' % (self.template, self.tenant)

    @staticmethod
    def get_quota_to_component_mapping():
        return {
            openstack_models.Tenant.Quotas.ram: PackageComponent.Types.RAM,
            openstack_models.Tenant.Quotas.vcpu: PackageComponent.Types.CORES,
            openstack_models.Tenant.Quotas.storage: PackageComponent.Types.STORAGE,
        }

    def get_quota_usage(self):
        """
        Returns map:
        {
            'ram': ram_quota.usage,
            'cores': vcpu_quota.usage,
            'storage': storage_quota.usage,
        }
        """
        mapping = self.get_quota_to_component_mapping()
        usage = {}
        quotas = {quota.name: quota.usage for quota in self.tenant.quotas.all()}
        for quota, component_type in mapping.items():
            if quota.name in quotas:
                usage[component_type] = quotas[quota.name]
        return usage

    class Meta:
        verbose_name = _('OpenStack VPC package')
        verbose_name_plural = _('OpenStack VPC packages')



import logging

import requests
from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes import fields as ct_fields
from django.contrib.contenttypes import models as ct_models
from django.contrib.postgres.fields import JSONField as BetterJSONField
from django.core import validators
from django.core.mail import send_mail
from django.db import models
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.lru_cache import lru_cache
from django.utils.translation import ugettext_lazy as _
from model_utils.fields import AutoCreatedField
from model_utils.models import TimeStampedModel

from waldur_core.core.fields import JSONField, UUIDField
from waldur_core.core.managers import GenericKeyMixin

logger = logging.getLogger(__name__)


class UuidMixin(models.Model):
    # There is circular dependency between logging and core applications.
    # Core models are loggable. So we cannot use UUID mixin here.

    class Meta:
        abstract = True

    uuid = UUIDField()


class AlertThresholdMixin(models.Model):
    """
    It is expected that model has scope field.
    """

    class Meta:
        abstract = True

    threshold = models.FloatField(default=0, validators=[validators.MinValueValidator(0)])

    def is_over_threshold(self):
        """
        If returned value is True, alert is generated.
        """
        raise NotImplementedError

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        from django.apps import apps
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @classmethod
    def get_checkable_objects(cls):
        """
        It should return queryset of objects that should be checked.
        """
        return cls.objects.all()


class EventTypesMixin(models.Model):
    """
    Mixin to add a event_types and event_groups fields.
    """

    class Meta:
        abstract = True

    event_types = BetterJSONField('List of event types')
    event_groups = BetterJSONField('List of event groups', default=list)

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]


class BaseHook(EventTypesMixin, UuidMixin, TimeStampedModel):
    class Meta:
        abstract = True

    user = models.ForeignKey(on_delete=models.CASCADE, to=settings.AUTH_USER_MODEL)
    is_active = models.BooleanField(default=True)

    # This timestamp would be updated periodically when event is sent via this hook
    last_published = models.DateTimeField(default=timezone.now)

    @property
    def all_event_types(self):
        from waldur_core.logging import loggers

        self_types = set(self.event_types)
        try:
            hook_ct = ct_models.ContentType.objects.get_for_model(self)
            base_types = SystemNotification.objects.get(hook_content_type=hook_ct)
        except SystemNotification.DoesNotExist:
            return self_types
        else:
            return self_types | set(loggers.expand_event_groups(base_types.event_groups)) | set(base_types.event_types)

    @classmethod
    def get_active_hooks(cls):
        return [obj for hook in cls.__subclasses__() for obj in hook.objects.filter(is_active=True)]

    @classmethod
    @lru_cache(maxsize=1)
    def get_all_models(cls):
        return [model for model in apps.get_models() if issubclass(model, cls)]

    @classmethod
    def get_all_content_types(cls):
        ctypes = ct_models.ContentType.objects.get_for_models(*cls.get_all_models())
        ids = [ctype.id for ctype in ctypes.values()]
        return ct_models.ContentType.objects.filter(id__in=ids)


class WebHook(BaseHook):
    class ContentTypeChoices:
        JSON = 1
        FORM = 2
        CHOICES = ((JSON, 'json'), (FORM, 'form'))

    destination_url = models.URLField()
    content_type = models.SmallIntegerField(
        choices=ContentTypeChoices.CHOICES,
        default=ContentTypeChoices.JSON
    )

    def process(self, event):
        logger.debug('Submitting web hook to URL %s, payload: %s', self.destination_url, event)
        payload = dict(
            created=event.created.isoformat(),
            message=event.message,
            context=event.context,
            event_type=event.event_type,
        )

        # encode event as JSON
        if self.content_type == WebHook.ContentTypeChoices.JSON:
            requests.post(self.destination_url, json=payload, verify=settings.VERIFY_WEBHOOK_REQUESTS)

        # encode event as form
        elif self.content_type == WebHook.ContentTypeChoices.FORM:
            requests.post(self.destination_url, data=payload, verify=settings.VERIFY_WEBHOOK_REQUESTS)


class PushHook(BaseHook):
    class Type:
        IOS = 1
        ANDROID = 2
        CHOICES = ((IOS, 'iOS'), (ANDROID, 'Android'))

    class Meta:
        unique_together = 'user', 'device_id', 'type'

    type = models.SmallIntegerField(choices=Type.CHOICES)
    device_id = models.CharField(max_length=255, null=True, unique=True)
    device_manufacturer = models.CharField(max_length=255, null=True, blank=True)
    device_model = models.CharField(max_length=255, null=True, blank=True)
    token = models.CharField(max_length=255, null=True, unique=True)

    def process(self, event):
        """ Send events as push notification via Google Cloud Messaging.
            Expected settings as follows:

                # https://developers.google.com/mobile/add
                WALDUR_CORE['GOOGLE_API'] = {
                    'NOTIFICATION_TITLE': "Waldur notification",
                    'Android': {
                        'server_key': 'AIzaSyA2_7UaVIxXfKeFvxTjQNZbrzkXG9OTCkg',
                    },
                    'iOS': {
                        'server_key': 'AIzaSyA34zlG_y5uHOe2FmcJKwfk2vG-3RW05vk',
                    }
                }
        """

        conf = settings.WALDUR_CORE.get('GOOGLE_API') or {}
        keys = conf.get(dict(self.Type.CHOICES)[self.type])

        if not keys or not self.token:
            return

        endpoint = 'https://gcm-http.googleapis.com/gcm/send'
        headers = {
            'Content-Type': 'application/json',
            'Authorization': 'key=%s' % keys['server_key'],
        }
        payload = {
            'to': self.token,
            'notification': {
                'body': event.message or 'New event',
                'title': conf.get('NOTIFICATION_TITLE', 'Waldur notification'),
                'image': 'icon',
            },
            'data': {
                'event': event.context
            },
        }
        if self.type == self.Type.IOS:
            payload['content-available'] = '1'
        logger.debug('Submitting GCM push notification with headers %s, payload: %s' % (headers, payload))
        requests.post(endpoint, json=payload, headers=headers)


class EmailHook(BaseHook):
    email = models.EmailField(max_length=75)

    def process(self, event):
        if not self.email:
            logger.debug('Skipping processing of email hook (PK=%s) because email is not defined' % self.pk)
            return
        subject = settings.WALDUR_CORE.get('NOTIFICATION_SUBJECT', 'Notifications from Waldur')
        text_message = event.message
        html_message = render_to_string('logging/email.html', {'events': [event]})
        logger.debug('Submitting email hook to %s, payload: %s', self.email, text_message)
        send_mail(subject, text_message, settings.DEFAULT_FROM_EMAIL, [self.email], html_message=html_message)


class SystemNotification(EventTypesMixin, models.Model):
    # Model doesn't inherit NameMixin, because this is circular dependence.
    name = models.CharField(_('name'), max_length=150)
    hook_content_type = models.ForeignKey(on_delete=models.CASCADE, to=ct_models.ContentType, related_name='+')
    roles = JSONField('List of roles', default=list)

    @staticmethod
    def get_valid_roles():
        return 'admin', 'manager', 'owner'

    @classmethod
    def get_hooks(cls, event_type, project=None, customer=None):
        from waldur_core.structure import models as structure_models
        from waldur_core.logging import loggers

        groups = [g[0] for g in loggers.event_logger.get_all_groups().items() if event_type in g[1]]

        for hook in cls.objects.filter(models.Q(event_types__contains=event_type) |
                                       models.Q(event_groups__has_any_keys=groups)):
            hook_class = hook.hook_content_type.model_class()
            users_qs = []

            if project:
                if 'admin' in hook.roles:
                    users_qs.append(project.get_users(structure_models.ProjectRole.ADMINISTRATOR))
                if 'manager' in hook.roles:
                    users_qs.append(project.get_users(structure_models.ProjectRole.MANAGER))
                if 'owner' in hook.roles:
                    users_qs.append(project.customer.get_owners())

            if customer:
                if 'owner' in hook.roles:
                    users_qs.append(customer.get_owners())

            if len(users_qs) > 1:
                users = users_qs[0].union(*users_qs[1:]).distinct()
            elif len(users_qs) == 1:
                users = users_qs[0]
            else:
                users = []

            for user in users:
                if user.email:
                    yield hook_class(
                        user=user,
                        event_types=hook.event_types,
                        email=user.email
                    )

    def __str__(self):
        return '%s | %s' % (self.hook_content_type, self.name)


class Report(UuidMixin, TimeStampedModel):
    class States:
        PENDING = 'pending'
        DONE = 'done'
        ERRED = 'erred'

        CHOICES = (
            (PENDING, 'Pending'),
            (DONE, 'Done'),
            (ERRED, 'Erred'),
        )

    file = models.FileField(upload_to='logging_reports')
    file_size = models.PositiveIntegerField(null=True)
    state = models.CharField(choices=States.CHOICES, default=States.PENDING, max_length=10)
    error_message = models.TextField(blank=True)


class Event(UuidMixin):
    created = AutoCreatedField()
    event_type = models.CharField(max_length=100, db_index=True)
    message = models.TextField()
    context = BetterJSONField(blank=True)

    class Meta:
        ordering = ('-created',)


class FeedManager(GenericKeyMixin, models.Manager):
    pass


class Feed(models.Model):
    event = models.ForeignKey(on_delete=models.CASCADE, to=Event)
    content_type = models.ForeignKey(on_delete=models.CASCADE, to=ct_models.ContentType, db_index=True)
    object_id = models.PositiveIntegerField(db_index=True)
    scope = ct_fields.GenericForeignKey('content_type', 'object_id')
    objects = FeedManager()
